import os, traceback, re, apsw, base64, io, qrcode
from nicegui import app, ui
from functools import partial

from biblemateweb import config, BIBLEMATEWEB_APP_DIR, getBibleVersionList, get_translation

from biblemateweb.pages.ai.chat import ai_chat
from biblemateweb.pages.ai.partner import ai_partner
from biblemateweb.pages.ai.partner_cli import ai_partner_cli
from biblemateweb.pages.ai.agent import ai_agent
from biblemateweb.pages.ai.agent_cli import ai_agent_cli

from agentmake.plugins.uba.lib.BibleBooks import BibleBooks
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
from biblemateweb.fx.bible_selection_dialog import BibleSelectionDialog

from biblemateweb.js.bible import BIBLE_JS
from biblemateweb.js.original import get_original_js

# Import for page content
from biblemateweb.pages.bibles.original_reader import original_reader
from biblemateweb.pages.bibles.original_interlinear import original_interlinear
from biblemateweb.pages.bibles.original_parallel import original_parallel
from biblemateweb.pages.bibles.original_discourse import original_discourse
from biblemateweb.pages.bibles.original_linguistic import original_linguistic
from biblemateweb.pages.bibles.bible_translation import bible_translation

from biblemateweb.pages.tools.xrefs import xrefs
from biblemateweb.pages.tools.treasury import treasury
from biblemateweb.pages.tools.audio import bibles_audio
from biblemateweb.pages.tools.podcast import bibles_podcast
from biblemateweb.pages.tools.commentary import bible_commentary
from biblemateweb.pages.tools.chronology import bible_chronology
from biblemateweb.pages.tools.timelines import bible_timelines
from biblemateweb.pages.tools.indexes import resource_indexes
from biblemateweb.pages.tools.chapter_indexes import chapter_indexes
from biblemateweb.pages.tools.promises import bible_promises_menu
from biblemateweb.pages.tools.parallels import bible_parallels_menu
from biblemateweb.pages.tools.morphology import word_morphology
from biblemateweb.pages.tools.notepad import notepad
from biblemateweb.pages.tools.notes import notes
from biblemateweb.pages.tools.summary import chapter_summary
from biblemateweb.pages.tools.analysis import book_analysis

from biblemateweb.pages.search.bible_verses import search_bible_verses
from biblemateweb.pages.search.bible_promises import search_bible_promises
from biblemateweb.pages.search.bible_parallels import search_bible_parallels
from biblemateweb.pages.search.bible_topics import search_bible_topics
from biblemateweb.pages.search.bible_locations import search_bible_locations
from biblemateweb.pages.search.bible_characters import search_bible_characters
from biblemateweb.pages.search.bible_names import search_bible_names
from biblemateweb.pages.search.dictionaries import search_bible_dictionaries
from biblemateweb.pages.search.encyclopedias import search_bible_encyclopedias
from biblemateweb.pages.search.lexicons import search_bible_lexicons
from biblemateweb.pages.search.bible_maps import search_bible_maps
from biblemateweb.pages.search.bible_relationships import search_bible_relationships

from biblemateweb.pages.teachings.parousia import parousia
from biblemateweb.pages.teachings.parousia_zh import parousia_zh

class BibleMateWeb:
    def __init__(self):

        # TODO: Consider dark mode latter
        # Dark Mode
        # ui.dark_mode().toggle()

        # Global variable to track current layout
        self.current_layout = app.storage.user['layout']
        self.area1_wrapper = None
        self.area2_wrapper = None
        self.splitter = None
        self.is_portrait = False

        # Tab panels and active tab tracking
        self.area1_tabs = None
        self.area2_tabs = None
        self.area1_tab_panels = {}  # Dictionary to store tab panels by name
        self.area2_tab_panels = {}
        self.area1_tab_panels_container = None
        self.area2_tab_panels_container = None

        # Tab number
        self.area1_tab_loaded = self.area1_tab_counter = 0
        self.area2_tab_loaded = self.area2_tab_counter = 0

        # tools
        self.tools = {
            "analysis": book_analysis,
            "summary": chapter_summary,
            "notes": notes,
            "notepad": notepad,
            "parousia": parousia,
            "parousia_zh": parousia_zh,
            "chat": ai_chat,
            "partner": ai_partner_cli if config.disable_partner_mode and not app.storage.user["api_key"].strip() else ai_partner,
            "agent": ai_agent_cli if config.disable_agent_mode and not app.storage.user["api_key"].strip() else ai_agent,
            "morphology": word_morphology,
            "indexes": resource_indexes,
            "chapterindexes": chapter_indexes,
            "podcast": bibles_podcast,
            "audio": bibles_audio,
            "verses": search_bible_verses, # API with additional options
            "treasury": treasury,
            "commentary": bible_commentary, # API with additional options
            "chronology": bible_chronology,
            "timelines": bible_timelines,
            "xrefs": xrefs,
            "promises": search_bible_promises,
            "promises_": bible_promises_menu,
            "parallels": search_bible_parallels,
            "parallels_": bible_parallels_menu,
            "topics": search_bible_topics,
            "characters": search_bible_characters,
            "locations": search_bible_locations,
            "names": search_bible_names,
            "dictionaries": search_bible_dictionaries,
            "encyclopedias": search_bible_encyclopedias, # API with additional options
            "lexicons": search_bible_lexicons, # API with additional options
            "maps": search_bible_maps,
            "relationships": search_bible_relationships,
        }

    def work_in_progress(self, **_):
        with ui.column().classes('w-full items-center'):
            ui.label('BibleMate AI').classes('text-2xl mt-4')
            ui.label('This feature is currently in progress.').classes('text-gray-600')
            ui.notify("This feature is currently in progress.")

    def check_breakpoint(self, ev):
        # prefer the well-known attributes
        # width
        width = getattr(ev, 'width', None)
        # fallback: some versions wrap data inside an attribute (try common names)
        if width is None:
            for maybe in ('args', 'arguments', 'data', 'payload'):
                candidate = getattr(ev, maybe, None)
                if isinstance(candidate, dict) and 'width' in candidate:
                    width = candidate['width']
                    break
        if width is None:
            print('Could not determine width from event:', ev)
            return
        # height
        height = getattr(ev, 'height', None)
        # fallback: some versions wrap data inside an attribute (try common names)
        if height is None:
            for maybe in ('args', 'arguments', 'data', 'payload'):
                candidate = getattr(ev, maybe, None)
                if isinstance(candidate, dict) and 'height' in candidate:
                    height = candidate['height']
                    break
        if height is None:
            print('Could not determine height from event:', ev)
            return
        self.is_portrait = width < height
        if self.splitter:
            if self.is_portrait:
                self.splitter.props('horizontal')
            else:
                self.splitter.props(remove='horizontal')

    def create_home_layout(self):
        """Create two scrollable areas with responsive layout"""
        
        # listen to the resize event
        ui.on('resize', self.check_breakpoint)
        
        # Inject JS
        ui.add_head_html('''
        <style>
            /* Make tab content area horizontal so close button is inline with label */
            .closable-tab .q-tab__content {
                flex-direction: row !important;
                align-items: center !important;
                gap: 2px;
            }
            /* Reduce tab padding for compactness */
            .closable-tab {
                padding-right: 6px !important;
            }
            /* Smaller close button */
            .closable-tab .close-btn {
                min-height: 18px !important;
                min-width: 18px !important;
                padding: 0 !important;
                margin-right: -4px !important;
            }
            .closable-tab .close-btn .q-icon {
                font-size: 12px !important;
            }
        </style>
        ''')
        ui.add_head_html(BIBLE_JS) # for active verse scrolling
        ui.add_head_html(get_original_js(app.storage.user['dark_mode'])) # for interactive highlighting

        # Create self.splitter
        self.splitter = ui.splitter(value=100, horizontal=self.is_portrait).classes('w-full').style('height: 100vh')

        # Area 1
        with self.splitter.before:
            previous_tabs1 = sorted([i for i in app.storage.user.keys() if i.startswith("tab1_")])
            default_number_of_tabs1 = app.storage.user.get("default_number_of_tabs1", 3)

            self.area1_wrapper = ui.column().classes('w-full h-full !gap-0')
            with self.area1_wrapper:
                self.area1_tabs = ui.tabs().classes('w-full')
                with self.area1_tabs:
                    if previous_tabs1:
                        for i in range(1, len(previous_tabs1)+1):
                            tab_id = f'tab1_{i}'
                            with ui.tab(f'tab1_{i}', label=app.storage.user.get(previous_tabs1[i-1]).get("label", f'Bible {i}')).classes('text-secondary closable-tab') as tab:
                                close_btn = ui.button(
                                    icon='close',
                                    on_click=partial(self.remove_tab_area1_any, tab_id),
                                ).props('flat dense round size=xs').classes('close-btn opacity-50 hover:opacity-100')
                                close_btn.on('click', js_handler='(e) => e.stopPropagation()')
                            self.area1_tab_counter += 1
                        self.area1_tab_loaded = self.area1_tab_counter
                    if len(previous_tabs1) < default_number_of_tabs1:
                        for i in range(len(previous_tabs1)+1, default_number_of_tabs1+1):
                            tab_id = f'tab1_{i}'
                            with ui.tab(tab_id, label=f'Bible {i}').classes('text-secondary closable-tab') as tab:
                                close_btn = ui.button(
                                    icon='close',
                                    on_click=partial(self.remove_tab_area1_any, tab_id),
                                ).props('flat dense round size=xs').classes('close-btn opacity-50 hover:opacity-100')
                                close_btn.on('click', js_handler='(e) => e.stopPropagation()')
                            self.area1_tab_counter += 1
                
                self.area1_tab_panels_container = ui.tab_panels(self.area1_tabs, value='tab1_1').classes('w-full h-full')
                
                with self.area1_tab_panels_container:

                    if previous_tabs1:
                        for i in range(1, len(previous_tabs1)+1):
                            tab_id = f'tab1_{i}'
                            saved_tab_id = previous_tabs1[i-1]
                            with ui.tab_panel(tab_id).classes('w-full h-full !p-0 !b-0 !m-0 !gap-0'):
                                self.area1_tab_panels[tab_id] = ui.scroll_area().classes(f'w-full h-full {tab_id}')
                                with self.area1_tab_panels[tab_id]:
                                    args = app.storage.user.get(saved_tab_id)
                                    content = self.get_content(args.get("title"))
                                    if content is None and saved_tab_id in app.storage.user:
                                        app.storage.user.pop(saved_tab_id)
                                        continue
                                    args["tab1"] = tab_id
                                    content(gui=self, **args)
                                    if saved_tab_id != tab_id and saved_tab_id in app.storage.user:
                                        app.storage.user[tab_id] = app.storage.user.pop(saved_tab_id)
                    if len(previous_tabs1) < default_number_of_tabs1:
                        for i in range(len(previous_tabs1)+1, default_number_of_tabs1+1):
                            tab_id = f'tab1_{i}'
                            with ui.tab_panel(tab_id).classes('w-full h-full !p-0 !b-0 !m-0 !gap-0'):
                                self.area1_tab_panels[tab_id] = ui.scroll_area().classes(f'w-full h-full {tab_id}')
                                with self.area1_tab_panels[tab_id]:
                                    ui.label(f'Bible Area - Tab {i}').classes('text-2xl font-bold mb-4')
                                    ui.label('Please select a Bible from the menu.').classes('text-gray-600')
        
        # Area 2
        with self.splitter.after:
            previous_tabs2 = sorted([i for i in app.storage.user.keys() if i.startswith("tab2_")])
            default_number_of_tabs2 = app.storage.user.get("default_number_of_tabs2", 3)

            self.area2_wrapper = ui.column().classes('w-full h-full !gap-0')
            with self.area2_wrapper:
                self.area2_tabs = ui.tabs().classes('w-full')
                with self.area2_tabs:
                    if previous_tabs2:
                        for i in range(1, len(previous_tabs2)+1):
                            tab_id = f'tab2_{i}'
                            with ui.tab(f'tab2_{i}', label=app.storage.user.get(previous_tabs2[i-1]).get("label", f'Tool {i}')).classes('text-secondary closable-tab') as tab:
                                close_btn = ui.button(
                                    icon='close',
                                    on_click=partial(self.remove_tab_area2_any, tab_id),
                                ).props('flat dense round size=xs').classes('close-btn opacity-50 hover:opacity-100')
                                close_btn.on('click', js_handler='(e) => e.stopPropagation()')
                            self.area2_tab_counter += 1
                        self.area2_tab_loaded = self.area2_tab_counter
                    if len(previous_tabs2) < default_number_of_tabs2:
                        for i in range(len(previous_tabs2)+1, default_number_of_tabs2+1):
                            tab_id = f'tab2_{i}'
                            with ui.tab(tab_id, label=f'Tool {i}').classes('text-secondary closable-tab') as tab:
                                close_btn = ui.button(
                                    icon='close',
                                    on_click=partial(self.remove_tab_area2_any, tab_id),
                                ).props('flat dense round size=xs').classes('close-btn opacity-50 hover:opacity-100')
                                close_btn.on('click', js_handler='(e) => e.stopPropagation()')
                            self.area2_tab_counter += 1
                
                self.area2_tab_panels_container = ui.tab_panels(self.area2_tabs, value='tab2_1').classes('w-full h-full')
                
                with self.area2_tab_panels_container:

                    if previous_tabs2:
                        for i in range(1, len(previous_tabs2)+1):
                            tab_id = f'tab2_{i}'
                            saved_tab_id = previous_tabs2[i-1]
                            with ui.tab_panel(tab_id).classes('w-full h-full !p-0 !b-0 !m-0 !gap-0'):
                                self.area2_tab_panels[tab_id] = ui.scroll_area().classes(f'w-full h-full {tab_id}')
                                with self.area2_tab_panels[tab_id]:
                                    args = app.storage.user.get(previous_tabs2[i-1])
                                    content = self.get_content(args.get("title"))
                                    if content is None and saved_tab_id in app.storage.user:
                                        app.storage.user.pop(saved_tab_id)
                                        continue
                                    args["tab2"] = tab_id
                                    content(gui=self, **args)
                                    if saved_tab_id != tab_id and saved_tab_id in app.storage.user:
                                        app.storage.user[tab_id] = app.storage.user.pop(saved_tab_id)
                    if len(previous_tabs2) < default_number_of_tabs2:
                        for i in range(len(previous_tabs2)+1, default_number_of_tabs2+1):
                            tab_id = f'tab2_{i}'
                            with ui.tab_panel(tab_id).classes('w-full h-full !p-0 !b-0 !m-0 !gap-0'):
                                self.area2_tab_panels[tab_id] = ui.scroll_area().classes(f'w-full h-full {tab_id}')
                                with self.area2_tab_panels[tab_id]:
                                    ui.label(f'Tool Area - Tab {i}').classes('text-2xl font-bold mb-4')
                                    ui.label('Please select a tool from the menu.').classes('text-gray-600')

        # A Draggable Container
        def make_draggable(element):
            """
            Applies drag-and-drop logic to any UI element.
            Snap-to-position on drag end.
            """
            def handle_drag_end(e):
                x = e.args['clientX']
                y = e.args['clientY']
                # Position element based on drop location, centered
                element.style(f'top: {y - 20}px; left: {x - 20}px; bottom: auto; right: auto')

            element.props('draggable="true"') \
                .style('cursor: grab') \
                .on('dragend', handle_drag_end, ['clientX', 'clientY'])
            return element

        with ui.column().classes('fixed bottom-6 right-6 z-50 touch-none') \
                .props('draggable="true"') \
                .style('cursor: grab') as self.fab_container1:
            make_draggable(self.fab_container1)
            self.fab_container1.bind_visibility_from(app.storage.user, 'layout_swap_button')

            # - 'fab-mini': Quasar's specific prop for a smaller floating action button.
            ui.button(icon='swap_horiz', on_click=self.swap_layout) \
                .props('fab-mini color=primary') \
                .tooltip(get_translation("Swap Layout"))

        BOOK_NAMES = {i: BibleBooks.abbrev[app.storage.user['ui_language']][str(i)][0] for i in range(1,67)}
        VERSES = BibleBooks.verses
        bible_selection_dialog = BibleSelectionDialog(self, BOOK_NAMES, VERSES)

        with ui.column().classes('fixed bottom-6 right-20 z-50 touch-none') \
                .props('draggable="true"') \
                .style('cursor: grab') as self.fab_container2:
            make_draggable(self.fab_container2)
            self.fab_container2.bind_visibility_from(app.storage.user, 'bible_select_button')

            # - 'fab-mini': Quasar's specific prop for a smaller floating action button.
            ui.button(icon='menu_book', on_click=bible_selection_dialog.open) \
                .props('fab-mini color=primary') \
                .tooltip(get_translation("Select Bible Verse"))

        # Set initial visibility
        self.update_visibility()

    def swap_layout(self, layout=None):
        """Swap between three layout modes"""
        if not layout in (1, 2, 3, None):
            layout = 2
        app.storage.user['layout'] = self.current_layout = layout if layout else (self.current_layout % 3) + 1
        self.update_visibility()

    def update_visibility(self):
        """Update visibility of areas based on current layout"""
        
        if self.current_layout == 1:
            # Area 1 visible, Area 2 invisible - maximize Area 1
            self.area1_wrapper.set_visibility(True)
            self.area2_wrapper.set_visibility(False)
            self.splitter.set_value(100)  # Move self.splitter to maximize Area 1
        elif self.current_layout == 2:
            # Both areas visible - 50/50 split
            self.area1_wrapper.set_visibility(True)
            self.area2_wrapper.set_visibility(True)
            self.splitter.set_value(50)  # Move self.splitter to middle
        elif self.current_layout == 3:
            # Area 1 invisible, Area 2 visible - maximize Area 2
            self.area1_wrapper.set_visibility(False)
            self.area2_wrapper.set_visibility(True)
            self.splitter.set_value(0)  # Move self.splitter to maximize Area 2

    def get_active_area1_tab(self):
        """Get the currently active tab in Area 1"""
        return self.area1_tab_panels_container.value

    def get_active_area2_tab(self):
        """Get the currently active tab in Area 2"""
        return self.area2_tab_panels_container.value

    def select_next_area1_tab(self, add_tab=True):
        if len(self.area1_tab_panels) == 1:
            if add_tab:
                self.add_tab_area1()
            else:
                return
        else:
            next_tab = False
            while not next_tab:
                for i in self.area1_tab_panels:
                    if next_tab:
                        self.area1_tab_panels_container.value = i
                        return None
                    elif i == self.area1_tab_panels_container.value:
                        next_tab = True

    def select_next_area2_tab(self, add_tab=True):
        if len(self.area2_tab_panels) == 1:
            if add_tab:
                self.add_tab_area2()
            else:
                return
        else:
            next_tab = False
            while not next_tab:
                for i in self.area2_tab_panels:
                    if next_tab:
                        self.area2_tab_panels_container.value = i
                        return None
                    elif i == self.area2_tab_panels_container.value:
                        next_tab = True

    def select_empty_area1_tab(self):
        for child in self.area1_tabs:
            if hasattr(child, '_props') and re.search("^Bible [0-9]", child._props.get('label', '')):
                if tab_name := child._props.get('name', ''):
                   self.area1_tab_panels_container.value = tab_name
                   return None
        self.add_tab_area1()

    def select_empty_area2_tab(self):
        for child in self.area2_tabs:
            if hasattr(child, '_props') and re.search("^Tool [0-9]", child._props.get('label', '')):
                if tab_name := child._props.get('name', ''):
                   self.area2_tab_panels_container.value = tab_name
                   return None
        self.add_tab_area2()

    def close_other_area1_tabs(self):
        keep_tab = self.area1_tab_panels_container.value
        for i in list(self.area1_tab_panels.keys()):
            if i != keep_tab:
                self.area1_tab_panels_container.value = i
                self.remove_tab_area1()

    def close_other_area2_tabs(self):
        keep_tab = self.area2_tab_panels_container.value
        for i in list(self.area2_tab_panels.keys()):
            if i != keep_tab:
                self.area2_tab_panels_container.value = i
                self.remove_tab_area2()

    def is_tool(self, title):
        return True if title.lower() in self.tools else False

    def get_content(self, title):
        if not isinstance(title, str):
            title = "NET"
        if title.lower() in self.tools:
            return self.tools[title.lower()]
        elif title == "ORB":
            return original_reader
        elif title == "OIB":
            return original_interlinear
        elif title == "OPB":
            return original_parallel
        elif title == "ODB":
            return original_discourse
        elif title == "OLB":
            return original_linguistic
        elif app.storage.client["custom"] and title in config.bibles_custom:
            return bible_translation
        elif title in config.bibles:
            return bible_translation
        else:
            return None

    def generate_qr_base64(self, data: str) -> str:
        """
        Generates a QR code for the given string and returns it 
        as a base64 encoded data URL for use in ui.image().
        """
        qr = qrcode.QRCode(
            version=1,
            error_correction=qrcode.constants.ERROR_CORRECT_L,
            box_size=10,
            border=4,
        )
        qr.add_data(data)
        qr.make(fit=True)
        
        img = qr.make_image(fill_color="black", back_color="white")
        
        # Save image to a memory buffer (avoiding file creation)
        buffer = io.BytesIO()
        img.save(buffer, format="PNG")
        img_str = base64.b64encode(buffer.getvalue()).decode()
        
        # Return standard data URL format
        return f'data:image/png;base64,{img_str}'

    def copy_text(self, text=""):
        if text:
            # Escape quotes and newlines for JavaScript
            escaped = text.replace('\\', '\\\\').replace('`', '\\`').replace('$', '\\$')
            ui.run_javascript(f'navigator.clipboard.writeText(`{escaped}`)')
            ui.notify(f'Copied: "{text}"')
        else:
            ui.notify('No text selected', type='warning')

    def add_to_notepad(self, context):
        app.storage.user["tool_query"] = context
        self.select_empty_area2_tab()
        self.load_area_2_content(title="note", sync=False)

    def ask_biblemate(self, context):
        app.storage.user["tool_query"] = context
        self.select_empty_area2_tab()
        self.load_area_2_content(title="chat", sync=False)

    def get_aic(self):
        if app.storage.user['ui_language'] == "tc":
            return "AICTC"
        elif app.storage.user['ui_language'] == "sc":
            return "AICSC"
        else:
            return "AIC"

    def open_book_context_menu(self, db, b, c, v, note=False):
        app.storage.user["tool_book_number"] = b
        app.storage.user["tool_chapter_number"] = c
        app.storage.user["tool_verse_number"] = v
        def open_tool(title):
            self.select_empty_area2_tab()
            self.load_area_2_content(title=title, sync=False)
        def open_analysis(section):
            app.storage.user["tool_query"] = section
            open_tool("Analysis")
        def open_book_note():
            app.storage.user["tool_chapter_number"] = 0
            app.storage.user["tool_verse_number"] = 0
            open_tool("Notes")
        with ui.context_menu() as menu:
            ui.menu_item(f'⏳ {get_translation("Timelines")}', on_click=lambda: open_tool("Timelines"))
            ui.separator()
            ui.menu_item(f'🔘 {get_translation("Overview")}', on_click=lambda: open_analysis("Overview"))
            ui.menu_item(f'📐 {get_translation("Structural Outline")}', on_click=lambda: open_analysis("Structural Outline"))
            ui.menu_item(f'🧵 {get_translation("Logical Flow")}', on_click=lambda: open_analysis("Logical Flow"))
            ui.menu_item(f'🕰️ {get_translation("Historical Setting")}', on_click=lambda: open_analysis("Historical Setting"))
            ui.menu_item(f'🎨 {get_translation("Themes")}', on_click=lambda: open_analysis("Themes"))
            ui.menu_item(f'🏷️ {get_translation("Keywords")}', on_click=lambda: open_analysis("Keywords"))
            ui.menu_item(f'⛪ {get_translation("Theology")}', on_click=lambda: open_analysis("Theology"))
            ui.menu_item(f'🚩 {get_translation("Canonical Placement")}', on_click=lambda: open_analysis("Canonical Placement"))
            ui.menu_item(f'🌱 {get_translation("Practical Living")}', on_click=lambda: open_analysis("Practical Living"))
            ui.menu_item(f'💎 {get_translation("Summary")}', on_click=lambda: open_analysis("Summary"))
            if config.google_client_id and config.google_client_secret:
                ui.separator()
                ui.menu_item(f'📝 {get_translation("Edit Note" if note else "Add Note")}', on_click=open_book_note)
        menu.open()

    def open_chapter_context_menu(self, db, b, c, v, note=False):
        app.storage.user["tool_book_number"] = b
        app.storage.user["tool_chapter_number"] = c
        app.storage.user["tool_verse_number"] = v
        def get_chapter_content():
            nonlocal b, c, v
            ref = BibleVerseParser(False, language=app.storage.user['ui_language']).bcvToVerseReference(b, c, v, isChapter=True)
            with apsw.Connection(db) as connn:
                query = "SELECT Verse, Scripture FROM Verses WHERE Book=? AND Chapter=? ORDER BY Verse"
                cursor = connn.cursor()
                cursor.execute(query, (b, c))
                fetches = cursor.fetchall()
            if not fetches: return ref
            verses = [f"[{verse}] {scripture}" for verse, scripture in fetches]
            return f"# {ref}\n\n" + "\n".join(verses)
        def open_tool(title):
            self.select_empty_area2_tab()
            self.load_area_2_content(title=title, sync=False)
        def open_map():
            nonlocal b, c, v
            ref = BibleVerseParser(False, language=app.storage.user['ui_language']).bcvToVerseReference(b, c, v, isChapter=True)
            app.storage.user["tool_query"] = f"{ref}:1-180"
            open_tool("Maps")
        def open_chapter_note():
            app.storage.user["tool_verse_number"] = 0
            open_tool("Notes")
        with ui.context_menu() as menu:
            ui.menu_item(f'📋 {get_translation("Copy")}', on_click=lambda: self.copy_text(get_chapter_content()))
            ui.separator()
            ui.menu_item(f'📡 {get_translation("Bible Podcast")}', on_click=lambda: open_tool("Podcast"))
            ui.menu_item(f'🔊 {get_translation("Bible Audio")}', on_click=lambda: open_tool("Audio"))
            ui.separator()
            ui.menu_item(f'💎 {get_translation("Chapter Summary")}', on_click=lambda: open_tool("Summary"))
            ui.menu_item(f'🗺️ {get_translation("Chapter Map")}', on_click=open_map)
            ui.menu_item(f'📑 {get_translation("Chapter Indexes")}', on_click=lambda: open_tool("Chapterindexes"))
            if config.google_client_id and config.google_client_secret:
                ui.separator()
                ui.menu_item(f'📝 {get_translation("Edit Note" if note else "Add Note")}', on_click=open_chapter_note)
        menu.open()

    def open_verse_context_menu(self, db, b, c, v, note=False):
        app.storage.user["tool_book_number"] = b
        app.storage.user["tool_chapter_number"] = c
        app.storage.user["tool_verse_number"] = v
        def get_verse_content():
            nonlocal b, c, v
            ref = BibleVerseParser(False, language=app.storage.user['ui_language']).bcvToVerseReference(b, c, v)
            with apsw.Connection(db) as connn:
                query = "SELECT Scripture FROM Verses WHERE Book=? AND Chapter=? AND Verse =?"
                cursor = connn.cursor()
                cursor.execute(query, (b, c, v))
                fetch = cursor.fetchone()
            verse_text = re.sub("<[^<>]+?>", "", fetch[0]) if fetch else ""
            return f"[{ref}] {verse_text}"
        def open_tool(title):
            self.select_empty_area2_tab()
            self.load_area_2_content(title=title, sync=False)
        def compare_verse():
            nonlocal db, b, c, v
            ref = BibleVerseParser(False, language=app.storage.user['ui_language']).bcvToVerseReference(b, c, v)
            bible_versions = sorted(list(set([app.storage.user["primary_bible"], app.storage.user["primary_bible"], os.path.basename(db)[:-6], "OHGBi"])))
            app.storage.user["tool_query"] = f"{','.join(bible_versions)}:::{ref}"
            open_tool("Verses")
        def add_to_notepad():
            nonlocal self
            self.add_to_notepad(get_verse_content())
        def ask_biblemate():
            nonlocal self
            ref, verse_content = get_verse_content()[1:].split("] ", 1)
            self.ask_biblemate(f"# {ref}\n\n{verse_content}\n\n# Query\n\n")
        with ui.context_menu() as menu:
            ui.menu_item(f'📋 {get_translation("Copy")}', on_click=lambda: self.copy_text(get_verse_content()))
            ui.separator()
            ui.menu_item(f'🔊 {get_translation("Bible Audio")}', on_click=lambda: open_tool("Audio"))
            ui.separator()
            ui.menu_item(f'🔗 {get_translation("Cross-references")}', on_click=lambda: open_tool("Xrefs"))
            ui.menu_item(f'🏦 {get_translation("Treasury")}', on_click=lambda: open_tool("Treasury"))
            ui.menu_item(f'🧠 {get_translation("AI Commentary")}', on_click=lambda: (
                app.storage.user.update(favorite_commentary=self.get_aic()),
                open_tool("Commentary")
            ))
            ui.menu_item(f'📚 {get_translation("Commentaries")}', on_click=lambda: open_tool("Commentary"))
            ui.separator()
            ui.menu_item(f'🧬 {get_translation("Morphology")}', on_click=lambda: open_tool("Morphology"))
            ui.separator()
            ui.menu_item(f'👀 {get_translation("Comparison")}', on_click=compare_verse)
            ui.menu_item(f'📑 {get_translation("Indexes")}', on_click=lambda: open_tool("Indexes"))
            ui.separator()
            if config.google_client_id and config.google_client_secret:
                ui.menu_item(f'📝 {get_translation("Edit Note" if note else "Add Note")}', on_click=lambda: open_tool("Notes"))
            else:
                ui.menu_item(f'📝 {get_translation("Add Note")}', on_click=add_to_notepad)
            ui.menu_item(f'💬 {get_translation("Ask BibleMate")}', on_click=ask_biblemate)
        menu.open()

    async def replace_url(self):
        new_url = ""
        active_bible_tab = self.get_active_area1_tab()
        if active_bible_tab in app.storage.user:
            args = app.storage.user[active_bible_tab]
            new_url = f'/?bbt={args.get("bt")}&bb={args.get("b")}&bc={args.get("c")}&bv={args.get("v")}'
        active_tool_tab = self.get_active_area2_tab()
        if active_tool_tab in app.storage.user:
            args = app.storage.user[active_tool_tab]
            title = args.get("bt")
            if not new_url:
                new_url = "/?"
            else:
                new_url += "&"
            new_url += f'tool={title if title.lower() in self.tools else "bible"}&tbt={args.get("bt")}&tb={args.get("b")}&tc={args.get("c")}&tv={args.get("v")}&tq={args.get("q")}'
        if new_url:
            new_url += f"&l={app.storage.user['layout']}"
        await ui.run_javascript(f"window.history.replaceState({{}}, '', '{new_url}')")

    def load_area_1_content(self, content=None, title="Bible", tab=None, args=None, keep=True, update_url=True):
        """Load example content in the active tab of Area 1"""

        if app.storage.user['layout'] == 3:
            self.swap_layout(2)

        if content is None:
            content = self.get_content(title)
            if content is None:
                print("No content found!")
                return None

        try:
            # modify tab label here for particular features TODO
            tab_label = title
            if not app.storage.user['bible_book_text'] == title:
                app.storage.user['bible_book_text'] = title
            # Get the currently active tab
            active_tab = tab if tab else self.get_active_area1_tab()
            # args holder
            args = args if args else {
                "title": title,
                "label": tab_label,
                "bt": app.storage.user.get('bible_book_text', app.storage.user["primary_bible"]),
                "b": app.storage.user.get('bible_book_number', 1),
                "c": app.storage.user.get('bible_chapter_number', 1),
                "v": app.storage.user.get('bible_verse_number', 1),
                "q": app.storage.user.get('bible_query', ''),
                "area": 1,
                "tab1": active_tab,
                "tab2": self.get_active_area2_tab(),
            }
            # store as history
            if update_url:
                with self.splitter: # attach to splitter to workaround `RuntimeError: The parent element this slot belongs to has been deleted.`
                    if client := ui.context.client:
                        new_url = f'/?bbt={args.get("bt")}&bb={args.get("b")}&bc={args.get("c")}&bv={args.get("v")}'
                        client.run_javascript(f"window.history.pushState({{}}, '', '{new_url}')")
                        ref = BibleVerseParser(False, language=app.storage.user['ui_language']).bcvToVerseReference(args.get("b"), args.get("c"), args.get("v"))
                        client.run_javascript(f'document.title = "[{title}] {ref}"')
            if keep:
                app.storage.user[active_tab] = args
            # Get the active tab's scroll area
            active_panel = self.area1_tab_panels[active_tab]
            # Clear and load new content
            active_panel.clear()
            # load content
            with active_panel:
                content(gui=self, **args)
            # Update tab label to reflect new content
            for child in self.area1_tabs:
                if hasattr(child, '_props') and child._props.get('name') == active_tab:
                    child.props(f'label="{tab_label}"')
                    break
            # reset bible query
            app.storage.user["bible_query"] = ""
        except:
            print(traceback.format_exc())

    def load_area_2_content(self, content=None, title="Tool", tab=None, args=None, keep=True, sync=True, update_url=True):
        """Load example content in the active tab of Area 2"""

        if app.storage.user['layout'] == 1:
            self.swap_layout(2)

        is_sync = True if sync or (sync and app.storage.user.get("sync") and not self.is_tool(title)) else False

        if content is None:
            content = self.get_content(title)
            if content is None:
                print("No content found!")
                return None

        try:
            # modify tab label here for particular features TODO
            tab_label = title
            # Get the currently active tab
            bible_tab = self.get_active_area1_tab()
            active_tab = tab if tab else self.get_active_area2_tab()
            # args holder
            sync_b = app.storage.user[bible_tab]["b"] if bible_tab in app.storage.user else app.storage.user.get('bible_book_number')
            sync_c = app.storage.user[bible_tab]["c"] if bible_tab in app.storage.user else app.storage.user.get('bible_chapter_number')
            sync_v = app.storage.user[bible_tab]["v"] if bible_tab in app.storage.user else app.storage.user.get('bible_verse_number')
            args = args if args else {
                "title": title,
                "label": tab_label,
                "bt": app.storage.user.get('tool_book_text'),
                "b": sync_b if is_sync else app.storage.user.get('tool_book_number'),
                "c": sync_c if is_sync else app.storage.user.get('tool_chapter_number'),
                "v": sync_v if is_sync else app.storage.user.get('tool_verse_number'),
                "q": app.storage.user.get('tool_query', ''),
                "area": 2,
                "tab1": bible_tab,
                "tab2": active_tab,
            }
            # store as history
            if update_url:
                with self.splitter: # attach to splitter to workaround `RuntimeError: The parent element this slot belongs to has been deleted.`
                    if client := ui.context.client:
                        new_url = f'/?tool={title if title.lower() in self.tools else "bible"}&tbt={args.get("bt")}&tb={args.get("b")}&tc={args.get("c")}&tv={args.get("v")}&tq={args.get("q")}'
                        client.run_javascript(f"window.history.pushState({{}}, '', '{new_url}')")
                        ref = BibleVerseParser(False, language=app.storage.user['ui_language']).bcvToVerseReference(args.get("b"), args.get("c"), args.get("v"))
                        client.run_javascript(f'document.title = "[{title.upper()}] {ref}"')
            if keep:
                app.storage.user[active_tab] = args
            # Get the active tab's scroll area
            active_panel = self.area2_tab_panels[active_tab]
            # Clear and load new content
            active_panel.clear()
            # load content
            with active_panel:
                content(gui=self, **args)
            # Update tab label to reflect new content
            for child in self.area2_tabs:
                if hasattr(child, '_props') and child._props.get('name') == active_tab:
                    child.props(f'label="{tab_label}"')
                    break
            # reset tool query
            app.storage.user["tool_query"] = ""
        except:
            print(traceback.format_exc())

    def update_active_area1_tab_records(self, title=None, label=None, bt=None, b=None, c=None, v=None, q=None):
        active_area1_tab = self.get_active_area1_tab()
        if active_area1_tab in app.storage.user: # only update when keep = True
            args = app.storage.user[active_area1_tab]
            if title is not None:
                args["title"] = title
            if label is not None:
                args["label"] = label
            if bt is not None:
                args["bt"] = bt
            if b is not None:
                args["b"] = b
            if c is not None:
                args["c"] = c
            if v is not None:
                args["v"] = v
            if q is not None:
                args["q"] = q
            app.storage.user[active_area1_tab] = args
        app.storage.user["bible_query"] = ""

    def update_active_area2_tab_records(self, title=None, label=None, bt=None, b=None, c=None, v=None, q=None):
        active_area2_tab = self.get_active_area2_tab()
        if active_area2_tab in app.storage.user: # only update when keep = True
            args = app.storage.user[active_area2_tab]
            if title is not None:
                args["title"] = title
            if label is not None:
                args["label"] = label
            if bt is not None:
                args["bt"] = bt
            if b is not None:
                args["b"] = b
            if c is not None:
                args["c"] = c
            if v is not None:
                args["v"] = v
            if q is not None:
                args["q"] = q
            app.storage.user[active_area2_tab] = args
        app.storage.user["tool_query"] = ""

    def get_area_1_bible_text(self):
        active_bible_tab = self.get_active_area1_tab()
        return app.storage.user[active_bible_tab]["bt"] if active_bible_tab in app.storage.user else app.storage.user['bible_book_text'] if 'bible_book_text' in app.storage.user else app.storage.user["primary_bible"]

    def get_area_2_bible_text(self):
        active_bible_tab = self.get_active_area2_tab()
        return app.storage.user[active_bible_tab]["bt"] if active_bible_tab in app.storage.user else app.storage.user['tool_book_text'] if 'tool_book_text' in app.storage.user else app.storage.user["primary_bible"]

    def show_all_verses(self, entry):
        #ui.notify('Loading all verses...')
        app.storage.user["tool_query"] = entry
        self.select_empty_area2_tab()
        self.load_area_2_content(title='Verses')

    def open_area_1_bible_bible(self, bible):
        if refs := BibleVerseParser(False).extractAllReferences(app.storage.user['tool_query']): # to work with text selection
            b, c, v, *_ = refs[0]
            self.change_area_1_bible_chapter(version=bible, book=b, chapter=c, verse=v)
        else:
            self.load_area_1_content(title=bible)

    def change_area_1_bible_chapter(self, version=None, book=1, chapter=1, verse=1):
        if version is None:
            version = self.get_area_1_bible_text()
        app.storage.user['bible_book_text'] = version
        app.storage.user['bible_book_number'] = book
        app.storage.user['bible_chapter_number'] = chapter
        app.storage.user['bible_verse_number'] = verse
        # in some cases, self.get_content do not work
        if version == "ORB":
            content = original_reader
        elif version == "OIB":
            content = original_interlinear
        elif version == "OPB":
            content = original_parallel
        elif version == "ODB":
            content = original_discourse
        elif version == "OLB":
            content = original_linguistic
        else:
            content = bible_translation
        self.load_area_1_content(content=content, title=version)
        if app.storage.user["sync"]:
            args = app.storage.user[self.get_active_area2_tab()]
            if not self.is_tool(args.get("title")) and (not args.get("b") == book or not args.get("c") == chapter):
                self.change_area_2_bible_chapter(args.get("bt"), book, chapter, verse)

    def change_area_2_bible_chapter(self, version=None, book=1, chapter=1, verse=1, sync=True):
        if version is None:
            version = self.get_area_2_bible_text()
        app.storage.user['tool_book_text'] = version
        if sync and app.storage.user.get("sync"):
            app.storage.user['bible_book_number'] = book
            app.storage.user['bible_chapter_number'] = chapter
            app.storage.user['bible_verse_number'] = verse
        else:
            app.storage.user['tool_book_number'] = book
            app.storage.user['tool_chapter_number'] = chapter
            app.storage.user['tool_verse_number'] = verse
        # in some cases, self.get_content do not work
        if version == "ORB":
            content = original_reader
        elif version == "OIB":
            content = original_interlinear
        elif version == "OPB":
            content = original_parallel
        elif version == "ODB":
            content = original_discourse
        elif version == "OLB":
            content = original_linguistic
        else:
            content = bible_translation
        self.load_area_2_content(content=content, title=version, sync=sync)
        if sync and app.storage.user["sync"]:
            args = app.storage.user[self.get_active_area1_tab()]
            if not args.get("b") == book or not args.get("c") == chapter:
                self.change_area_1_bible_chapter(args.get("bt"), book, chapter, verse)

    def add_tab_area1(self):
        """Dynamically add a new tab to Area 1"""
        self.area1_tab_counter += 1
        new_tab_name = f'tab1_{self.area1_tab_counter}'
        # Add new tab
        with self.area1_tabs:
            with ui.tab(new_tab_name, label=f'Bible {self.area1_tab_counter}').classes('text-secondary closable-tab') as tab:
                close_btn = ui.button(
                    icon='close',
                    on_click=partial(self.remove_tab_area1_any, new_tab_name),
                ).props('flat dense round size=xs').classes('close-btn opacity-50 hover:opacity-100')
                close_btn.on('click', js_handler='(e) => e.stopPropagation()')
        # Add new tab panel
        with self.area1_tab_panels_container:
            with ui.tab_panel(new_tab_name).classes('w-full h-full !p-0 !b-0 !m-0 !gap-0'):
                self.area1_tab_panels[new_tab_name] = ui.scroll_area().classes(f'w-full h-full {new_tab_name}')
                with self.area1_tab_panels[new_tab_name]:
                    ui.label(f'Bible Area - Tab {self.area1_tab_counter}').classes('text-2xl font-bold mb-4')
                    ui.label('Please select a Bible from the menu.').classes('text-gray-600')
        self.area1_tabs.set_value(new_tab_name)

    def remove_tab_area1_any(self, id):
        """Remove any tabs from Area 1, even they are not active"""
        active_tab = self.area1_tab_panels_container.value
        if not active_tab == id:
            self.area1_tab_panels_container.value = id
        self.remove_tab_area1()
        if not active_tab == id and not self.area1_tab_panels_container.value == active_tab:
            self.area1_tab_panels_container.value = active_tab

    def remove_tab_area1(self):
        """Remove the currently active tab from Area 1"""
        active_tab = self.get_active_area1_tab()
        # Don't allow removing if it's the last tab
        if len(self.area1_tab_panels) <= 1:
            ui.notify('Cannot remove the last tab!', type='warning')
            return
        # Find and remove the tab
        tab_to_remove = None
        for child in self.area1_tabs:
            if hasattr(child, '_props') and child._props.get('name') == active_tab:
                tab_to_remove = child
                break
        if tab_to_remove:
            # Switch to a different tab before removing
            remaining_tabs = [k for k in self.area1_tab_panels.keys() if k != active_tab]
            if remaining_tabs:
                self.area1_tab_panels_container.set_value(remaining_tabs[0])
            # Remove the tab
            self.area1_tabs.remove(tab_to_remove)
            # Remove the tab panel
            if active_tab in self.area1_tab_panels:
                self.area1_tab_panels[active_tab].parent_slot.parent.delete()
                del self.area1_tab_panels[active_tab]
            if active_tab in app.storage.user:
                del app.storage.user[active_tab]
        #self.area1_tab_counter = len(self.area1_tab_panels) # do not update, otherwise new tab records may override the existing ones

    def add_tab_area2(self):
        """Dynamically add a new tab to Area 2"""
        self.area2_tab_counter += 1
        new_tab_name = f'tab2_{self.area2_tab_counter}'
        # Add new tab
        with self.area2_tabs:
            with ui.tab(new_tab_name, label=f'Tool {self.area2_tab_counter}').classes('text-secondary closable-tab') as tab:
                close_btn = ui.button(
                    icon='close',
                    on_click=partial(self.remove_tab_area2_any, new_tab_name),
                ).props('flat dense round size=xs').classes('close-btn opacity-50 hover:opacity-100')
                close_btn.on('click', js_handler='(e) => e.stopPropagation()')
        # Add new tab panel
        with self.area2_tab_panels_container:
            with ui.tab_panel(new_tab_name).classes('w-full h-full !p-0 !b-0 !m-0 !gap-0'):
                self.area2_tab_panels[new_tab_name] = ui.scroll_area().classes(f'w-full h-full {new_tab_name}')
                with self.area2_tab_panels[new_tab_name]:
                    ui.label(f'Tool Area - Tab {self.area2_tab_counter}').classes('text-2xl font-bold mb-4')
                    ui.label('Please select a tool from the menu.').classes('text-gray-600')
        self.area2_tabs.set_value(new_tab_name)

    def remove_tab_area2_any(self, id):
        """Remove any tabs from Area 2, even they are not active"""
        active_tab = self.area2_tab_panels_container.value
        if not active_tab == id:
            self.area2_tab_panels_container.value = id
        self.remove_tab_area2()
        if not active_tab == id and not self.area2_tab_panels_container.value == active_tab:
            self.area2_tab_panels_container.value = active_tab

    def remove_tab_area2(self):
        """Remove the currently active tab from Area 2"""
        active_tab = self.get_active_area2_tab()
        # Don't allow removing if it's the last tab
        if len(self.area2_tab_panels) <= 1:
            ui.notify('Cannot remove the last tab!', type='warning')
            return
        # Find and remove the tab
        tab_to_remove = None
        for child in self.area2_tabs:
            if hasattr(child, '_props') and child._props.get('name') == active_tab:
                tab_to_remove = child
                break
        if tab_to_remove:
            # Switch to a different tab before removing
            remaining_tabs = [k for k in self.area2_tab_panels.keys() if k != active_tab]
            if remaining_tabs:
                self.area2_tab_panels_container.set_value(remaining_tabs[0])
            # Remove the tab
            self.area2_tabs.remove(tab_to_remove)
            # Remove the tab panel
            if active_tab in self.area2_tab_panels:
                self.area2_tab_panels[active_tab].parent_slot.parent.delete()
                del self.area2_tab_panels[active_tab]
            if active_tab in app.storage.user:
                del app.storage.user[active_tab]
        #self.area2_tab_counter = len(self.area2_tab_panels) # do not update, otherwise new tab records may override the existing ones

    # --- Shared Menu Function ---
    # This function creates the header, horizontal menu (desktop),
    # and drawer (mobile).

    def create_menu(self):
        """Create the responsive header and navigation drawer."""

        auto_suggestions = [BibleBooks.abbrev[app.storage.user['ui_language']][str(i)][0] for i in range(1,67)]
        auto_suggestions += [f"{i}:::" for i in self.tools.keys()]

        parser = BibleVerseParser(False)
        def perform_quick_search(quick_search):
            app.storage.user.update(left_drawer_open=False)
            if search_item := quick_search.value.strip():
                client_bibles = getBibleVersionList(app.storage.client["custom"])
                refs = parser.extractAllReferences(search_item)
                if search_item.lower().startswith("bible:::") and refs: # open a parallel bible chapter in area 2
                    search_item = search_item[8:]
                    b,c,v, *_ = refs[0]
                    if ":::" in search_item and search_item.split(":::", 1)[0].strip() in client_bibles: # bible version is specified
                        version, search_item = search_item.split(":::", 1)
                        version = version.strip()
                    else: # bible version is not specified
                        version = None
                    self.change_area_2_bible_chapter(version=version, book=b, chapter=c, verse=v)
                elif ":::" in search_item and search_item.split(":::", 1)[0].strip().lower() in self.tools: # open a tool in area 2
                    tool, app.storage.user["tool_query"] = search_item.split(":::", 1)
                    tool = tool.strip()
                    self.load_area_2_content(title=tool, sync=app.storage.user["sync"])
                elif len(refs) == 1: # open a bible chapter in area 1
                    b,c,v, *_ = refs[0]
                    if ":::" in search_item and search_item.split(":::", 1)[0].strip() in client_bibles: # bible version is specified
                        search_item = search_item[8:]
                        version, search_item = search_item.split(":::", 1)
                        version = version.strip()
                    else: # bible version is not specified
                        version = None
                    self.change_area_1_bible_chapter(version=version,book=b, chapter=c, verse=v)
                else: # search for verses in area 2; when no reference or more than a reference
                    app.storage.user["tool_query"] = search_item
                    self.load_area_2_content(title='Verses', sync=app.storage.user["sync"])

        # qr code dialog
        with ui.dialog() as dialog, ui.card().classes('items-center text-center p-6'):
            ui.label('URL & QR Code').classes('text-xl font-bold text-secondary')
            
            # Container to hold dynamic content (URL label + QR image)
            qr_container = ui.column().classes('items-center gap-4')

            ui.link("[BibleMate AI]", "https://github.com/eliranwong/biblemate").classes('text-secondary break-all max-w-[300px]')
            
            ui.button('Close', on_click=dialog.close).props('outline align=center text-color=secondary')

        # --- Event Handler ---
        async def show_url_popup():
            await self.replace_url()

            # 1. Get the current URL from the user's browser
            # We must await this because it requires a round-trip to the client
            current_url = await ui.run_javascript('window.location.href')
            
            # 2. Update the dialog content
            qr_container.clear() # Clear previous content
            with qr_container:
                # Show URL (clickable link)
                ui.link(current_url, current_url).classes('break-all max-w-[300px]')
                
                # Show QR Code
                # We generate it on the fly based on the fetched URL
                base64_img = self.generate_qr_base64(current_url)
                ui.image(base64_img).style('width: 250px; height: 250px')
                
            # 3. Open the dialog
            dialog.open()

        # --- Header ---
        with ui.header(elevated=True).classes('bg-primary text-white p-0'):
            # We use 'justify-between' to push the left and right groups apart
            with ui.row().classes('w-full items-center justify-between no-wrap'):
                
                # --- Left Aligned Group ---
                with ui.row().classes('items-center no-wrap'):
                    # --- Hamburger Button (Mobile Only) ---
                    # This button toggles the 'left_drawer_open' value in user storage
                    # .classes('lt-sm') means "visible only on screens LESS THAN Medium"
                    ui.button(
                        on_click=lambda: app.storage.user.update(left_drawer_open=not app.storage.user['left_drawer_open']),
                        icon='menu'
                    ).props('flat color=white').classes('lt-sm')

                    # --- Desktop Avatar + Title (Home) ---
                    # The button contains a row with the avatar and the label
                    with ui.button(on_click=lambda: ui.timer(0, show_url_popup, once=True)).props('flat text-color=white').classes('gt-xs'):
                        with ui.row().classes('items-center no-wrap'):
                            # Use a fallback icon in case the image fails to load
                            with ui.avatar(size='32px'):
                                with ui.image(app.storage.user["avatar"] if app.storage.user["avatar"] else os.path.expanduser(config.avatar) if config.avatar else os.path.join(BIBLEMATEWEB_APP_DIR, 'eliranwong.jpg')) as image:
                                    with image.add_slot('error'):
                                        ui.icon('account_circle').classes('m-auto') # Center fallback icon
                            
                            # This is just a label now; the parent button handles the click
                            ui.label('BibleMate AI').classes('text-lg ml-2') # Added margin-left for spacing

                quick_search1 = ui.input(placeholder=f'🔍 {get_translation("Quick search")} ...', autocomplete=auto_suggestions) \
                        .props('clearable outlined rounded dense autofocus enterkeyhint="search"') \
                        .classes('gt-xs flex-grow')
                quick_search1.on('keydown.enter.prevent', lambda: perform_quick_search(quick_search1))

                # --- Right Aligned Group (Features & About Us) ---
                with ui.row().classes('items-center no-wrap'):

                    primary_bible = app.storage.user["primary_bible"]
                    secondary_bible = app.storage.user["secondary_bible"]
                    
                    #with ui.row().classes('gt-xs items-center overflow-x-auto overflow-y-hidden no-wrap'):                            
                    # Bibles
                    with ui.button(icon='local_library').props('flat color=white round').tooltip(get_translation("Bibles")):
                        with ui.menu():
                            ui.menu_item(get_translation("Add Bible Tab"), on_click=self.add_tab_area1)
                            ui.menu_item(get_translation("Close Bible Tab"), on_click=self.remove_tab_area1)
                            ui.menu_item(get_translation("Close Others"), on_click=self.close_other_area1_tabs)
                            ui.separator()
                            ui.menu_item(primary_bible, on_click=lambda: self.open_area_1_bible_bible(primary_bible)).tooltip(primary_bible)
                            ui.menu_item(secondary_bible, on_click=lambda: self.open_area_1_bible_bible(secondary_bible)).tooltip(secondary_bible)
                            ui.separator()
                            ui.menu_item(get_translation("Original Reader’s Bible"), on_click=lambda: self.open_area_1_bible_bible('ORB')).tooltip('ORB')
                            ui.menu_item(get_translation("Original Interlinear Bible"), on_click=lambda: self.open_area_1_bible_bible('OIB')).tooltip('OIB')
                            ui.menu_item(get_translation("Original Parallel Bible"), on_click=lambda: self.open_area_1_bible_bible('OPB')).tooltip('OPB')
                            ui.menu_item(get_translation("Original Discourse Bible"), on_click=lambda: self.open_area_1_bible_bible('ODB')).tooltip('ODB')
                            ui.menu_item(get_translation("Original Linguistic Bible"), on_click=lambda: self.open_area_1_bible_bible('OLB')).tooltip('OLB')
                            ui.separator()
                            if app.storage.client["custom"] and config.bibles_custom:
                                for i in config.bibles_custom:
                                    if not i in (primary_bible, secondary_bible):
                                        ui.menu_item(i, on_click=partial(self.open_area_1_bible_bible, i)).tooltip(config.bibles_custom[i][0])
                                ui.separator()
                            for i in config.bibles:
                                if not i in (primary_bible, secondary_bible) and ((app.storage.client["custom"] and not i in config.bibles_custom) or not app.storage.client["custom"]):
                                    ui.menu_item(i, on_click=partial(self.open_area_1_bible_bible, i)).tooltip(config.bibles[i][0])

                    with ui.button(icon='devices_fold').props('flat color=white round').tooltip(get_translation("Parallel Bibles")):
                        with ui.menu():
                            ui.menu_item(get_translation("Add Parallel Tab"), on_click=self.add_tab_area2)
                            ui.menu_item(get_translation("Close Parallel Tab"), on_click=self.remove_tab_area2)
                            ui.menu_item(get_translation("Close Others"), on_click=self.close_other_area2_tabs)
                            ui.separator()
                            ui.menu_item(primary_bible, on_click=lambda: self.load_area_2_content(title=primary_bible)).tooltip(primary_bible)
                            ui.menu_item(secondary_bible, on_click=lambda: self.load_area_2_content(title=secondary_bible)).tooltip(secondary_bible)
                            ui.separator()
                            ui.menu_item(get_translation("Original Reader’s Bible"), on_click=lambda: self.load_area_2_content(title='ORB')).tooltip('ORB')
                            ui.menu_item(get_translation("Original Interlinear Bible"), on_click=lambda: self.load_area_2_content(title='OIB')).tooltip('OIB')
                            ui.menu_item(get_translation("Original Parallel Bible"), on_click=lambda: self.load_area_2_content(title='OPB')).tooltip('OPB')
                            ui.menu_item(get_translation("Original Discourse Bible"), on_click=lambda: self.load_area_2_content(title='ODB')).tooltip('ODB')
                            ui.menu_item(get_translation("Original Linguistic Bible"), on_click=lambda: self.load_area_2_content(title='OLB')).tooltip('OLB')
                            ui.separator()
                            if app.storage.client["custom"] and config.bibles_custom:
                                for i in config.bibles_custom:
                                    if not i in (primary_bible, secondary_bible):
                                        ui.menu_item(i, on_click=partial(self.load_area_2_content, title=i)).tooltip(config.bibles_custom[i][0])
                                ui.separator()
                            for i in config.bibles:
                                if not i in (primary_bible, secondary_bible) and ((app.storage.client["custom"] and not i in config.bibles_custom) or not app.storage.client["custom"]):
                                    ui.menu_item(i, on_click=partial(self.load_area_2_content, title=i)).tooltip(config.bibles[i][0])
                            

                    # Bible Tools
                    with ui.button(icon='build').props('flat color=white round').tooltip(get_translation("Tools")):
                        with ui.menu():
                            ui.menu_item(get_translation("Add Tool Tab"), on_click=self.add_tab_area2)
                            ui.menu_item(get_translation("Close Tool Tab"), on_click=self.remove_tab_area2)
                            ui.menu_item(get_translation("Close Others"), on_click=self.close_other_area2_tabs)
                            ui.separator()
                            ui.menu_item(get_translation("Bible Podcast"), on_click=lambda: self.load_area_2_content(title='Podcast', sync=True))
                            ui.menu_item(get_translation("Bible Audio"), on_click=lambda: self.load_area_2_content(title='Audio', sync=True))
                            ui.separator()
                            ui.menu_item(get_translation("Book Analysis"), on_click=lambda: self.load_area_2_content(title='Analysis', sync=True))
                            ui.menu_item(get_translation("Chapter Summary"), on_click=lambda: self.load_area_2_content(title='Summary', sync=True))
                            ui.separator()
                            ui.menu_item(get_translation("Bible Commentaries"), on_click=lambda: self.load_area_2_content(title='Commentary', sync=True))
                            ui.menu_item(get_translation("Cross-references"), on_click=lambda: self.load_area_2_content(title='Xrefs', sync=True))
                            ui.menu_item(get_translation("Treasury of Scripture Knowledge"), on_click=lambda: self.load_area_2_content(title='Treasury', sync=True))
                            #ui.menu_item(get_translation("Discourse Analysis"), on_click=lambda: self.load_area_2_content(self.work_in_progress))
                            #ui.menu_item(get_translation("Morphological Data"), on_click=lambda: self.load_area_2_content(self.work_in_progress))
                            #ui.menu_item(get_translation("Translation Spectrum"), on_click=lambda: self.load_area_2_content(self.work_in_progress))
                            ui.separator()
                            ui.menu_item(get_translation("Bible Promises"), on_click=lambda: self.load_area_2_content(title='Promises_'))
                            ui.menu_item(get_translation("Bible Parallels"), on_click=lambda: self.load_area_2_content(title='Parallels_'))
                            ui.separator()
                            ui.menu_item(get_translation("Bible Timelines"), on_click=lambda: self.load_area_2_content(title='Timelines', sync=True))
                            ui.menu_item(get_translation("Bible Chronology"), on_click=lambda: self.load_area_2_content(title='Chronology'))
                            ui.separator()
                            ui.menu_item(get_translation("Morphology"), on_click=lambda: self.load_area_2_content(title='Morphology', sync=True))
                            ui.separator()
                            ui.menu_item(get_translation("Chapter Indexes"), on_click=lambda: self.load_area_2_content(title='Chapterindexes', sync=True))
                            ui.menu_item(get_translation("Verse Indexes"), on_click=lambda: self.load_area_2_content(title='Indexes', sync=True))
                            ui.separator()
                            if config.google_client_id and config.google_client_secret:
                                ui.menu_item(get_translation("Bible Notes"), on_click=lambda: self.load_area_2_content(title='Notes', sync=True))
                            ui.menu_item(get_translation("Notepad"), on_click=lambda: self.load_area_2_content(title='Notepad', sync=True))
                    
                    with ui.button(icon='search').props('flat color=white round').tooltip(get_translation("Search")):
                        with ui.menu():
                            #ui.menu_item(get_translation("Add Search Tab"), on_click=self.add_tab_area2)
                            #ui.menu_item(get_translation("Remove Search Tab"), on_click=self.remove_tab_area2)
                            #ui.menu_item(get_translation("Close Others"), on_click=self.close_other_area2_tabs)
                            #ui.separator()
                            ui.menu_item(get_translation("Verses"), on_click=lambda: self.load_area_2_content(title='Verses'))
                            ui.menu_item(get_translation("Parallels"), on_click=lambda: self.load_area_2_content(title='Parallels'))
                            ui.menu_item(get_translation("Promises"), on_click=lambda: self.load_area_2_content(title='Promises'))
                            ui.menu_item(get_translation("Topics"), on_click=lambda: self.load_area_2_content(title='Topics'))
                            ui.menu_item(get_translation("Characters"), on_click=lambda: self.load_area_2_content(title='Characters'))
                            ui.menu_item(get_translation("Relationships"), on_click=lambda: self.load_area_2_content(title='Relationships'))
                            ui.menu_item(get_translation("Locations"), on_click=lambda: self.load_area_2_content(title='Locations'))
                            ui.menu_item(get_translation("Maps"), on_click=lambda: self.load_area_2_content(title='Maps'))
                            ui.menu_item(get_translation("Names"), on_click=lambda: self.load_area_2_content(title='Names'))
                            ui.menu_item(get_translation("Dictionaries"), on_click=lambda: self.load_area_2_content(title='Dictionaries'))
                            ui.menu_item(get_translation("Encyclopedias"), on_click=lambda: self.load_area_2_content(title='Encyclopedias'))
                            ui.menu_item(get_translation("Lexicons"), on_click=lambda: self.load_area_2_content(title='Lexicons'))

                    with ui.button(icon='auto_awesome').props('flat color=white round').tooltip(get_translation("AI")):
                        with ui.menu():
                            ui.menu_item(get_translation("Book Analysis"), on_click=lambda: self.load_area_2_content(title='Analysis'))
                            ui.menu_item(get_translation("Chapter Summary"), on_click=lambda: self.load_area_2_content(title='Summary'))
                            ui.menu_item(get_translation("Commentary"), on_click=lambda: (
                                app.storage.user.update(favorite_commentary=self.get_aic()),
                                self.load_area_2_content(title='Commentary', sync=True)
                            ))
                            ui.separator()
                            #ui.menu_item('AI Q&A', on_click=lambda: self.load_area_2_content(self.work_in_progress))
                            ui.menu_item(get_translation("AI Chat"), on_click=lambda: self.load_area_2_content(title='Chat'))
                            ui.menu_item(get_translation("Partner Mode"), on_click=lambda: self.load_area_2_content(title='Partner'))
                            ui.menu_item(get_translation("Agent Mode"), on_click=lambda: self.load_area_2_content(title='Agent'))

                    with ui.button(icon='settings').props('flat color=white round').tooltip(get_translation("Settings")):
                        with ui.menu():
                            with ui.row().classes('w-full justify-between'):                            
                                # Back Button
                                ui.button(icon='arrow_back', on_click=lambda: ui.run_javascript('history.back()')) \
                                    .props('dense flat round') \
                                    .tooltip('Go Back')
                                #ui.space()
                                # Forward Button
                                ui.button(icon='arrow_forward', on_click=lambda: ui.run_javascript('history.forward()')) \
                                    .props('dense flat round') \
                                    .tooltip('Go Forward')
                            # swap layout
                            ui.menu_item(get_translation("Bible Only"), on_click=lambda: self.swap_layout(1))
                            ui.menu_item(get_translation("Tool Only"), on_click=lambda: self.swap_layout(3))
                            ui.menu_item(get_translation("Bible & Tool"), on_click=lambda: self.swap_layout(2))
                            # swap
                            def toggleSwapButton():
                                app.storage.user["layout_swap_button"] = not app.storage.user["layout_swap_button"]
                            with ui.row().tooltip(get_translation("Toggle Display of Swap Layout Button")):
                                ui.menu_item(get_translation("Swap"), on_click=toggleSwapButton)
                                ui.space()
                                ui.switch().bind_value(app.storage.user, 'layout_swap_button')
                            # navigate
                            def toggleBibleSelectionButton():
                                app.storage.user["bible_select_button"] = not app.storage.user["bible_select_button"]
                            with ui.row().tooltip(get_translation("Toggle Display of Bible Selection Button")):
                                ui.menu_item(get_translation("Go"), on_click=toggleBibleSelectionButton)
                                ui.space()
                                ui.switch().bind_value(app.storage.user, 'bible_select_button')
                            # notes
                            def toggleNotes():
                                app.storage.user["notes"] = not app.storage.user["notes"]
                                ui.run_javascript('location.reload()')
                            with ui.row().tooltip(get_translation("Toggle Note Indicators")):
                                ui.menu_item(get_translation("Notes"), on_click=toggleNotes)
                                ui.space()
                                ui.switch(value=app.storage.user["notes"], on_change=toggleNotes)
                            # sync
                            def toggleSync():
                                app.storage.user["sync"] = not app.storage.user["sync"]
                            with ui.row().tooltip(get_translation("Toggle Bible Synchronization")):
                                ui.menu_item(get_translation("Sync"), on_click=toggleSync)
                                ui.space()
                                ui.switch().bind_value(app.storage.user, 'sync')
                            ui.separator()
                            # full screen
                            def toggleFullscreen(): # ui.fullscreen().toggle does not work in this case
                                app.storage.user["fullscreen"] = not app.storage.user["fullscreen"]
                            with ui.row().tooltip(get_translation("Toggle Fullscreen")):
                                ui.menu_item(get_translation("Screen"), on_click=toggleFullscreen)
                                ui.space()
                                ui.switch().bind_value(app.storage.user, 'fullscreen')
                            # dark mode
                            with ui.menu_item() as dark_mode_menu_item:
                                dark_mode_label = ui.label(get_translation("Light Mode") if app.storage.user["dark_mode"] else get_translation("Dark Mode")).classes('flex items-center')
                            def toggle_dark_mode_menu_item(text_label: ui.label):
                                app.storage.user['dark_mode'] = not app.storage.user['dark_mode']
                                #text_label.set_text("Light Mode" if app.storage.user["dark_mode"] else "Dark Mode")
                                ui.run_javascript('location.reload()')
                            dark_mode_menu_item.on('click', lambda: toggle_dark_mode_menu_item(dark_mode_label))
                            ui.separator()
                            ui.menu_item(get_translation("Preferences"), on_click=lambda: ui.navigate.to('/settings'))

        # --- Drawer (Mobile Menu) ---
        # This section is unchanged
        with ui.drawer('left') \
                .classes('lt-sm') \
                .props('overlay') \
                .bind_value(app.storage.user, 'left_drawer_open') as left_drawer:

            # The button contains a row with the avatar and the label
            with ui.button(on_click=lambda: (
                ui.timer(0, show_url_popup, once=True),
                app.storage.user.update(left_drawer_open=False)
            )).props('flat text-color=white'):
                with ui.row().classes('items-center no-wrap'):
                    # Use a fallback icon in case the image fails to load
                    with ui.avatar(size='32px'):
                        with ui.image(app.storage.user["avatar"] if app.storage.user["avatar"] else os.path.expanduser(config.avatar) if config.avatar else os.path.join(BIBLEMATEWEB_APP_DIR, 'eliranwong.jpg')) as image:
                            with image.add_slot('error'):
                                ui.icon('account_circle').classes('m-auto') # Center fallback icon
                    
                    # This is just a label now; the parent button handles the click
                    ui.label('BibleMate AI').classes('text-lg ml-2')

            quick_search2 = ui.input(placeholder=f'🔍 {get_translation("Quick search")} ...', autocomplete=auto_suggestions) \
                    .props('clearable outlined rounded dense autofocus enterkeyhint="search" hide-bottom-space') \
                    .classes('w-full !m-0 !p-0')
            quick_search2.on('keydown.enter.prevent', lambda: perform_quick_search(quick_search2))

            ui.switch(get_translation("Go")).bind_value(app.storage.user, 'bible_select_button')
            ui.switch(get_translation("Swap")).bind_value(app.storage.user, 'layout_swap_button')
            ui.switch(get_translation("Notes"), value=app.storage.user["notes"], on_change=toggleNotes)
            ui.switch(get_translation("Sync")).bind_value(app.storage.user, 'sync')
            ui.switch(get_translation("Fullscreen")).bind_value(app.storage.user, 'fullscreen')
            ui.switch(get_translation("Dark Mode")).bind_value(app.storage.user, 'dark_mode').on_value_change(lambda: ui.run_javascript('location.reload()'))

            # Bibles
            with ui.expansion(get_translation("Bibles"), icon='local_library').props('header-class="text-secondary"'):
                ui.item(get_translation("Original Reader’s Bible"), on_click=lambda: (
                    self.open_area_1_bible_bible('ORB'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable').tooltip('ORB')
                ui.item(get_translation("Original Interlinear Bible"), on_click=lambda: (
                    self.open_area_1_bible_bible('OIB'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable').tooltip('OIB')
                ui.item(get_translation("Original Parallel Bible"), on_click=lambda: (
                    self.open_area_1_bible_bible('OPB'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable').tooltip('OPB')
                ui.item(get_translation("Original Discourse Bible"), on_click=lambda: (
                    self.open_area_1_bible_bible('ODB'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable').tooltip('ODB')
                ui.item(get_translation("Original Linguistic Bible"), on_click=lambda: (
                    self.open_area_1_bible_bible('OLB'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable').tooltip('OLB')
                ui.separator()
                if app.storage.client["custom"] and config.bibles_custom:
                    for i in config.bibles_custom:
                        ui.item(i, on_click=lambda: (
                            self.open_area_1_bible_bible(i),
                            app.storage.user.update(left_drawer_open=False)
                        )).props('clickable').tooltip(config.bibles_custom[i][0])
                    ui.separator()
                for i in config.bibles:
                    if (app.storage.client["custom"] and not i in config.bibles_custom) or not app.storage.client["custom"]:
                        ui.item(i, on_click=lambda: (
                            self.open_area_1_bible_bible(i),
                            app.storage.user.update(left_drawer_open=False)
                        )).props('clickable').tooltip(config.bibles[i][0])

            # Parallel Bibles
            with ui.expansion(get_translation("Parallel Bibles"), icon='devices_fold').props('header-class="text-secondary"'):
                ui.item(get_translation("Original Reader’s Bible"), on_click=lambda: (
                    self.load_area_2_content(title='ORB'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable').tooltip('ORB')
                ui.item(get_translation("Original Interlinear Bible"), on_click=lambda: (
                    self.load_area_2_content(title='OIB'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable').tooltip('OIB')
                ui.item(get_translation("Original Parallel Bible"), on_click=lambda: (
                    self.load_area_2_content(title='OPB'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable').tooltip('OPB')
                ui.item(get_translation("Original Discourse Bible"), on_click=lambda: (
                    self.load_area_2_content(title='ODB'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable').tooltip('ODB')
                ui.item(get_translation("Original Linguistic Bible"), on_click=lambda: (
                    self.load_area_2_content(title='OLB'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable').tooltip('OLB')
                ui.separator()
                if app.storage.client["custom"] and config.bibles_custom:
                    for i in config.bibles_custom:
                        ui.item(i, on_click=lambda: (
                            self.load_area_2_content(title=i),
                            app.storage.user.update(left_drawer_open=False)
                        )).props('clickable').tooltip(config.bibles_custom[i][0])
                    ui.separator()
                for i in config.bibles:
                    if (app.storage.client["custom"] and not i in config.bibles_custom) or not app.storage.client["custom"]:
                        ui.item(i, on_click=lambda: (
                            self.load_area_2_content(title=i),
                            app.storage.user.update(left_drawer_open=False)
                        )).props('clickable').tooltip(config.bibles[i][0])

            # Bible Tools
            with ui.expansion(get_translation("Tools"), icon='build').props('header-class="text-secondary"'):
                ui.item(get_translation("Bible Podcast"), on_click=lambda: (
                    self.load_area_2_content(title='Podcast', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Bible Audio"), on_click=lambda: (
                    self.load_area_2_content(title='Audio', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.separator()
                ui.item(get_translation("Book Analysis"), on_click=lambda: (
                    self.load_area_2_content(title='Analysis', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Chapter Summary"), on_click=lambda: (
                    self.load_area_2_content(title='Summary', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.separator()
                ui.item(get_translation("Bible Commentaries"), on_click=lambda: (
                    self.load_area_2_content(title='Commentary', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Cross-references"), on_click=lambda: (
                    self.load_area_2_content(title='Xrefs', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Treasury of Scripture Knowledge"), on_click=lambda: (
                    self.load_area_2_content(title='Treasury', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.separator()
                #ui.item(get_translation("Discourse Analysis"), on_click=lambda: (
                #    self.load_area_2_content(self.work_in_progress),
                #    app.storage.user.update(left_drawer_open=False)
                #)).props('clickable')
                #ui.item(get_translation("Morphological Data"), on_click=lambda: (
                #    self.load_area_2_content(self.work_in_progress),
                #    app.storage.user.update(left_drawer_open=False)
                #)).props('clickable')
                #ui.item(get_translation("Translation Spectrum"), on_click=lambda: (
                #    self.load_area_2_content(self.work_in_progress),
                #    app.storage.user.update(left_drawer_open=False)
                #)).props('clickable')
                ui.item(get_translation("Bible Promises"), on_click=lambda: (
                    self.load_area_2_content(title='Promises_'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Bible Parallels"), on_click=lambda: (
                    self.load_area_2_content(title='Parallels_'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.separator()
                ui.item(get_translation("Bible Timelines"), on_click=lambda: (
                    self.load_area_2_content(title='Timelines', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Bible Chronology"), on_click=lambda: (
                    self.load_area_2_content(title='Chronology'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.separator()
                ui.item(get_translation("Morphology"), on_click=lambda: (
                    self.load_area_2_content(title='Morphology', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.separator()
                ui.item(get_translation("Chapter Indexes"), on_click=lambda: (
                    self.load_area_2_content(title='Chapterindexes', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Verse Indexes"), on_click=lambda: (
                    self.load_area_2_content(title='Indexes', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.separator()
                if config.google_client_id and config.google_client_secret:
                    ui.item(get_translation("Bible Notes"), on_click=lambda: (
                        self.load_area_2_content(title='Notes', sync=True),
                        app.storage.user.update(left_drawer_open=False)
                    )).props('clickable')
                ui.item(get_translation("Notepad"), on_click=lambda: (
                    self.load_area_2_content(title='Notepad', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')

            # Search
            with ui.expansion(get_translation("Search"), icon='search').props('header-class="text-secondary"'):
                ui.item(get_translation("Verses"), on_click=lambda: (
                    self.load_area_2_content(title='Verses'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Parallels"), on_click=lambda: (
                    self.load_area_2_content(title='Parallels'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Promises"), on_click=lambda: (
                    self.load_area_2_content(title='Promises'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Topics"), on_click=lambda: (
                    self.load_area_2_content(title='Topics'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Characters"), on_click=lambda: (
                    self.load_area_2_content(title='Characters'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Relationships"), on_click=lambda: (
                    self.load_area_2_content(title='Relationships'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Locations"), on_click=lambda: (
                    self.load_area_2_content(title='Locations'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Maps"), on_click=lambda: (
                    self.load_area_2_content(title='Maps'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Names"), on_click=lambda: (
                    self.load_area_2_content(title='Names'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Dictionaries"), on_click=lambda: (
                    self.load_area_2_content(title='Dictionaries'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Encyclopedias"), on_click=lambda: (
                    self.load_area_2_content(title='Encyclopedias'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Lexicons"), on_click=lambda: (
                    self.load_area_2_content(title='Lexicons'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
            
            # AI
            with ui.expansion(get_translation("AI"), icon='auto_awesome').props('header-class="text-secondary"'):
                ui.item(get_translation("Book Analysis"), on_click=lambda: (
                    self.load_area_2_content(title='Analysis'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Chapter Summary"), on_click=lambda: (
                    self.load_area_2_content(title='Summary'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Commentary"), on_click=lambda: (
                    app.storage.user.update(favorite_commentary=self.get_aic()),
                    self.load_area_2_content(title='Commentary', sync=True),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.separator()
                #ui.item('AI Q&A', on_click=lambda: (
                #    self.load_area_2_content(self.work_in_progress),
                #    app.storage.user.update(left_drawer_open=False)
                #)).props('clickable')
                ui.item(get_translation("AI Chat"), on_click=lambda: (
                    self.load_area_2_content(title='Chat'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Partner Mode"), on_click=lambda: (
                    self.load_area_2_content(title='Partner'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Agent Mode"), on_click=lambda: (
                    self.load_area_2_content(title='Agent'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')

            # Preferences
            with ui.expansion(get_translation("Settings"), icon='settings').props('header-class="text-secondary"'):
                ui.item(get_translation("Bible Only"), on_click=lambda: (
                    self.swap_layout(1),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Tool Only"), on_click=lambda: (
                    self.swap_layout(2),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.item(get_translation("Bible & Tool"), on_click=lambda: (
                    self.swap_layout(3),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')
                ui.separator()
                ui.item(get_translation("Preferences"), on_click=lambda: (
                    ui.navigate.to('/settings'),
                    app.storage.user.update(left_drawer_open=False)
                )).props('clickable')

            with ui.row().classes('w-full justify-between'):                            
                # Back Button
                ui.button(color="secondary", icon='arrow_back', on_click=lambda: ui.run_javascript('history.back()')) \
                    .props('flat round') \
                    .tooltip('Go Back')
                # Forward Button
                ui.button(color="secondary", icon='arrow_forward', on_click=lambda: ui.run_javascript('history.forward()')) \
                    .props('flat round') \
                    .tooltip('Go Forward')


from nicegui import ui, app
from biblemateweb import config, getBibleVersionList, BIBLEMATEWEB_DATA
from typing import List, Optional
from agentmake.plugins.uba.lib.BibleBooks import BibleBooks
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
import re, apsw, os
from biblemate.api.bible import BibleVectorDatabase


def regexp(expr, item):
    reg = re.compile(expr, flags=0 if app.storage.user['search_case_sensitivity'] else re.IGNORECASE)
    return reg.search(item) is not None

def regexp_api(expr, item):
    reg = re.compile(expr, flags=re.IGNORECASE)
    return reg.search(item) is not None

def get_bible_content(user_input="", bible="NET", sql_query="", refs=[], search_mode=1, top_similar_verses=20, search_case_sensitivity=False, api=False, parser=None, html=True) -> list:
    verses_limit_reached = False
    dbs = []
    if isinstance(bible, str): # str; single bible
        if bible_path := getBiblePath(bible):
            dbs = [bible_path]
    else: # list; multiple bibles
        dbs = []
        for i in bible:
            if i_path := getBiblePath(i):
                if not i_path in dbs:
                    dbs.append(i_path)
    if not dbs:
        return []
    if parser is None:
        parser = BibleVerseParser(False)
    results = []
    if not refs and re.search(" [0-9]+?[:：][0-9]", user_input):
        refs = parser.extractAllReferences(user_input, tagged=(True if '<ref onclick="bcv(' in user_input else False))
    if not refs and search_mode == 3: # semantic search
        vector_db = BibleVectorDatabase(os.path.join(BIBLEMATEWEB_DATA, "vectors", "bible.db"))
        query = sql_query if sql_query else "PRAGMA case_sensitive_like = false; SELECT Book, Chapter, Verse, Scripture FROM Verses WHERE (Scripture REGEXP ?) ORDER BY Book, Chapter, Verse"
        if books := re.search("Book IN (.*?) AND ", query):
            book=books.group(1)
        elif books := re.search("Book=([0-9]+?) AND ", query):
            book=f"({books.group(1)})"
        else:
            book=0
        refs = [(b, c, v) for b, c, v, _ in vector_db.search_meaning(user_input, top_k=top_similar_verses, book=book)]
    if refs:
        distinct_refs = []
        for ref in refs:
            if ref not in distinct_refs:
                distinct_refs.append(ref)
        refs = distinct_refs
        if len(refs) > config.verses_limit:
            refs = refs[:config.verses_limit]
            verses_limit_reached = True
        query = "SELECT Scripture FROM Verses WHERE Book=? AND Chapter=? AND Verse =?"
        for db in dbs:
            this_bible = os.path.basename(db)[:-6]
            with apsw.Connection(db) as connn:
                cursor = connn.cursor()
                for ref in refs:
                    if len(ref) == 5:
                        content = ""
                        b1, c1, v1 = 1, 1, 1
                        for r in parser.extractExhaustiveReferences([ref]):
                            b, c, v = r
                            if not content:
                                b1, c1, v1 = r
                            cursor.execute(query, (b, c, v))
                            verse = cursor.fetchone()
                            if not verse: continue
                            verse_text = verse[0].strip()
                            if not html:
                                verse_text = re.sub("<.*?>", "", verse_text)
                            content += f"<vid>{v}</vid> {verse_text} "
                        ref = parser.bcvToVerseReference(*ref)
                        if len(dbs) > 1:
                            ref += f" [{this_bible}]"
                        results.append({'ref': ref, 'content': content.rstrip(), 'bible': this_bible, 'b': b1, 'c': c1, 'v': v1})
                    else:
                        b, c, v = ref
                        cursor.execute(query, (b, c, v))
                        verse = cursor.fetchone()
                        if not verse: continue
                        verse_text = verse[0].strip()
                        if not html:
                            verse_text = re.sub("<.*?>", "", verse_text)
                        ref = parser.bcvToVerseReference(b, c, v)
                        if len(dbs) > 1:
                            ref += f" [{this_bible}]"
                        results.append({'ref': ref, 'content': verse_text, 'bible': this_bible, 'b': b, 'c': c, 'v': v})
    else:
        # search the bible with regular expression
        query = sql_query if sql_query else "PRAGMA case_sensitive_like = false; SELECT Book, Chapter, Verse, Scripture FROM Verses WHERE (Scripture REGEXP ?) ORDER BY Book, Chapter, Verse"
        if search_case_sensitivity:
            query = query.replace("case_sensitive_like = false;", "case_sensitive_like = true;")
        if search_mode == 1:
            query = query.replace("Scripture REGEXP", "Scripture LIKE")
            user_input = "%"+user_input+"%"
        for db in dbs:
            this_bible = os.path.basename(db)[:-6]
            with apsw.Connection(db) as connn:
                connn.createscalarfunction("REGEXP", regexp_api if api else regexp)
                cursor = connn.cursor()
                cursor.execute(query, (user_input,))
                fetches = cursor.fetchall()
                if fetches and len(fetches) > config.verses_limit:
                    fetches = fetches[:config.verses_limit]
                    verses_limit_reached = True
                for verse in fetches:
                    verse_text = verse[3].strip()
                    if not html:
                        verse_text = re.sub("<.*?>", "", verse_text)
                    ref = parser.bcvToVerseReference(verse[0], verse[1], verse[2])
                    if len(dbs) > 1:
                        ref += f" [{this_bible}]"
                    results.append({'ref': ref, 'content': verse_text, 'bible': this_bible, 'b': verse[0], 'c': verse[1], 'v': verse[2]})
    if verses_limit_reached:
        results.append({'ref': "", 'content': f"{config.verses_limit} verses limit reached!", 'bible': "", 'b': 0, 'c': 0, 'v': 0})
    return results

# Bible Selection

def getBiblePath(bible) -> str:
    if bible in ["ORB", "OPB", "ODB", "OLB", "BHS5", "OGNT"]:
        bible = "OHGB"
    elif bible == "OIB":
        bible = "OHGBi"
    return config.bibles_custom[bible][-1] if bible in config.bibles_custom else config.bibles[bible][-1] if bible in config.bibles else ""

def getBibleChapter(db, b, c) -> str: # html output
    query = "SELECT Scripture FROM Bible WHERE Book=? AND Chapter=?"
    content = ""
    try:
        with apsw.Connection(db) as connn:
            #connn.createscalarfunction("REGEXP", regexp)
            cursor = connn.cursor()
            cursor.execute(query, (b, c))
            if scripture := cursor.fetchone():
                content = scripture[0]
    except:
        try:
            verses = [formatHTMLverse(i) for i in getBibleChapterVerses(db, b, c)]
            content = "<br>".join(verses)
        except Exception as e:
            content = "Error: "+str(e)
    return content

def formatHTMLverse(verse) -> str:
    b, c, v, text = verse
    return f"""<verse><vid id="v{b}.{c}.{v}" onclick="luV({v})">{v}</vid> {text}</verse>"""

def getBibleChapterVerses(db, b, c) -> str:
    query = "SELECT * FROM Verses WHERE Book=? AND Chapter=? ORDER BY Verse"
    verses = []
    with apsw.Connection(db) as connn:
        cursor = connn.cursor()
        cursor.execute(query, (b, c))
        verses = cursor.fetchall()
    return verses

def getBibleBookList(db) -> list:
    query = "SELECT DISTINCT Book FROM Verses ORDER BY Book"
    bookList = ""
    with apsw.Connection(db) as connn:
        cursor = connn.cursor()
        cursor.execute(query)
        bookList = sorted([book[0] for book in cursor.fetchall() if book[0]])
    return bookList

def getBibleChapterList(db, b) -> list:
    query = "SELECT DISTINCT Chapter FROM Verses WHERE Book=? ORDER BY Chapter"
    chapterList = ""
    with apsw.Connection(db) as connn:
        cursor = connn.cursor()
        cursor.execute(query, (b,))
        chapterList = sorted([chapter[0] for chapter in cursor.fetchall()])
    return chapterList

def getBibleVerseList(db, b, c) -> list:
    query = "SELECT DISTINCT Verse FROM Verses WHERE Book=? AND Chapter=? ORDER BY Verse"
    verseList = ""
    with apsw.Connection(db) as connn:
        cursor = connn.cursor()
        cursor.execute(query, (b, c))
        verseList = sorted([verse[0] for verse in cursor.fetchall()])
    return verseList

def change_bible_chapter_verse(_, book, chapter, verse):
    ui.run_javascript(f'scrollToVerse("v{book}.{chapter}.{verse}")')

class BibleSelector:
    """Class to manage Bible verse selection with dynamic dropdowns"""
    
    def __init__(self, on_version_changed=None, on_book_changed=None, on_chapter_changed=None, on_verse_changed=None, version_options=[], chapter_zero=False, verse_zero=False):
        # Handlers that replace the default on_change functions
        self.on_version_changed, self.on_book_changed, self.on_chapter_changed, self.on_verse_changed = on_version_changed, on_book_changed, on_chapter_changed, on_verse_changed

        # Initialize selected values
        self.selected_version: Optional[str] = None
        self.selected_book: Optional[str] = None
        self.selected_chapter: Optional[int] = None
        self.selected_verse: Optional[int] = None
        self.show_verses: Optional[bool] = None
        
        # Initialize dropdown UI elements
        self.version_select: Optional[ui.select] = None
        self.book_select: Optional[ui.select] = None
        self.chapter_select: Optional[ui.select] = None
        self.verse_select: Optional[ui.select] = None
        
        # Initialize options
        self.version_options: List[str] = version_options
        self.book_options: List[str] = []
        self.chapter_options: List[int] = []
        self.verse_options: List[int] = []

        self.chapter_zero, self.verse_zero = chapter_zero, verse_zero
        
    def create_ui(self, bible, b, c, v, additional_items=None, show_versions=True, show_verses=True):
        self.selected_version = bible
        self.selected_book = b
        self.selected_chapter = c
        self.selected_verse = v
        self.show_verses = show_verses

        if not self.version_options:
            self.version_options = getBibleVersionList(app.storage.client["custom"])
        bible_book_list = getBibleBookList(getBiblePath(self.selected_version))
        self.book_options = [BibleBooks.abbrev[app.storage.user['ui_language']][str(i)][0] for i in bible_book_list if str(i) in BibleBooks.abbrev[app.storage.user['ui_language']]]
        self.chapter_options = getBibleChapterList(getBiblePath(self.selected_version), self.selected_book)
        if self.chapter_zero:
            self.chapter_options.insert(0, 0)
        self.verse_options = [0] if self.chapter_zero and self.selected_chapter == 0 else getBibleVerseList(getBiblePath(self.selected_version), self.selected_book, self.selected_chapter)
        if self.verse_zero and not (self.chapter_zero and self.selected_chapter == 0):
            self.verse_options.insert(0, 0)
        try:
            default_book = BibleBooks.abbrev[app.storage.user['ui_language']][str(self.selected_book)][0]
        except:
            self.selected_book = bible_book_list[0]
            default_book = self.book_options[0]
            self.chapter_options = getBibleChapterList(getBiblePath(self.selected_version), self.selected_book)
            if self.chapter_zero:
                self.chapter_options.insert(0, 0)
            self.verse_options = [0] if self.chapter_zero and self.selected_chapter == 0 else getBibleVerseList(getBiblePath(self.selected_version), self.selected_book, self.selected_chapter)
            if self.verse_zero and not (self.chapter_zero and self.selected_chapter == 0):
                self.verse_options.insert(0, 0)
        with ui.row().classes('w-full justify-center items-center'):
            # Versions
            if show_versions:
                self.version_select = ui.select(
                    options=self.version_options,
                    #label='Bible',
                    value=bible if bible in self.version_options else self.version_options[0],
                    on_change=self.on_version_change
                )
            # Book
            self.book_select = ui.select(
                options=self.book_options,
                #label='Book',
                value=default_book if default_book in self.book_options else self.book_options[0], # b
                on_change=self.on_book_change
            )
            # Chapter
            self.chapter_select = ui.select(
                options=self.chapter_options,
                #label='Chapter',
                value=c if c in self.chapter_options else self.chapter_options[0],
                on_change=self.on_chapter_change
            )
            # Verse
            if show_verses:
                self.verse_select = ui.select(
                    options=self.verse_options,
                    #label='Verse',
                    value=v if v in self.verse_options else self.verse_options[0],
                    on_change=self.on_verse_change
                )
            if additional_items:
                additional_items()
    
    def on_version_change(self, e):
        """Handle Bible version selection change"""
        self.selected_version = e.value

        # replace default action
        if self.on_version_changed is not None:
            return self.on_version_changed(self.selected_version)
        
        self.reset_book_dropdown()
        self.reset_chapter_dropdown()
        self.reset_verse_dropdown()
    
    def on_book_change(self, e):
        """Handle book selection change"""
        self.selected_book = BibleBooks.bookNameToNum(e.value)

        # replace default action
        if self.on_book_changed is not None:
            return self.on_book_changed(self.selected_version, self.selected_book)

        self.reset_chapter_dropdown()
        self.reset_verse_dropdown()
    
    def on_chapter_change(self, e):
        """Handle chapter selection change"""
        self.selected_chapter = e.value

        # replace default action
        if self.on_chapter_changed is not None:
            return self.on_chapter_changed(self.selected_version, self.selected_book, self.selected_chapter)

        # Reset verse dropdown
        self.reset_verse_dropdown()
    
    def on_verse_change(self, e):
        """Handle verse selection change"""
        self.selected_verse = e.value

        # replace default action
        if self.on_verse_changed is not None:
            return self.on_verse_changed(self.selected_version, self.selected_book, self.selected_chapter, self.selected_verse)

    def reset_book_dropdown(self):
        """Reset book dropdown to initial state"""
        book_list = getBibleBookList(getBiblePath(self.selected_version))
        self.book_options = [BibleBooks.abbrev[app.storage.user['ui_language']][str(i)][0] for i in book_list if str(i) in BibleBooks.abbrev[app.storage.user['ui_language']]]
        self.book_select.options = self.book_options
        self.book_select.value = self.book_options[0]
        self.selected_book = book_list[0]
        # refresh
        self.book_select.update()
    
    def reset_chapter_dropdown(self):
        """Reset chapter dropdown to initial state"""
        self.chapter_options = getBibleChapterList(getBiblePath(self.selected_version), self.selected_book)
        if self.chapter_zero:
            self.chapter_options.insert(0, 0)
        self.chapter_select.options = self.chapter_options
        self.chapter_select.value = self.chapter_options[0]
        self.selected_chapter = self.chapter_options[0]
        # refresh
        self.chapter_select.update()
    
    def reset_verse_dropdown(self):
        """Reset verse dropdown to initial state"""
        if not self.show_verses:
            return
        self.verse_options = [0] if self.chapter_zero and self.selected_chapter == 0 else getBibleVerseList(getBiblePath(self.selected_version), self.selected_book, self.selected_chapter)
        if self.verse_zero and not (self.chapter_zero and self.selected_chapter == 0):
            self.verse_options.insert(0, 0)
        self.verse_select.options = self.verse_options
        self.verse_select.value = self.verse_options[0]
        self.selected_verse = self.verse_options[0]
        # refresh
        self.verse_select.update()
    
    def get_selection(self):
        """Get the current selection and display it"""
        return (self.selected_version, self.selected_book, self.selected_chapter, self.selected_verse)



TOOLTIP_JS = '''
<script>
    let currentTooltip = null;
    let tooltipCache = {};
    let tooltipTimeout = null;
    let currentWord = null;
    let showTimeout = null;
    let initialized = false;
    
    async function fetchTooltipData(word) {
        // Check cache first
        if (tooltipCache[word]) {
            return tooltipCache[word];
        }
        
        try {
            const response = await fetch(`/api/tooltip/${encodeURIComponent(word)}`);
            if (response.ok) {
                const data = await response.json();
                tooltipCache[word] = data;
                return data;
            }
        } catch (error) {
            console.error('Error fetching tooltip:', error);
        }
        return null;
    }
    
    function positionTooltip(tooltip, element) {
        // Get element and viewport dimensions
        const rect = element.getBoundingClientRect();
        const tooltipRect = tooltip.getBoundingClientRect();
        const viewportWidth = window.innerWidth;
        const viewportHeight = window.innerHeight;
        
        // Margins from viewport edges
        const margin = 10;
        const arrowSize = 10;
        
        // Calculate available space in each direction
        const spaceTop = rect.top;
        const spaceBottom = viewportHeight - rect.bottom;
        const spaceLeft = rect.left;
        const spaceRight = viewportWidth - rect.right;
        
        // Determine best position
        let position = 'top'; // default
        let top, left;
        
        // Remove old arrows
        const oldArrows = tooltip.querySelectorAll('.tooltip-arrow, .tooltip-arrow-border');
        oldArrows.forEach(arrow => arrow.remove());
        
        // Try to position tooltip
        if (spaceTop > tooltipRect.height + arrowSize + margin && spaceTop > spaceBottom) {
            // Position above
            position = 'top';
            top = rect.top - tooltipRect.height - arrowSize;
            left = rect.left + rect.width / 2 - tooltipRect.width / 2;
            
            // Adjust horizontal position if needed
            if (left < margin) {
                left = margin;
            } else if (left + tooltipRect.width > viewportWidth - margin) {
                left = viewportWidth - tooltipRect.width - margin;
            }
            
            // Add arrow pointing down
            const arrowBorder = document.createElement('div');
            arrowBorder.className = 'tooltip-arrow-border arrow-down';
            tooltip.appendChild(arrowBorder);
            
            const arrow = document.createElement('div');
            arrow.className = 'tooltip-arrow arrow-down';
            tooltip.appendChild(arrow);
            
            // Adjust arrow position if tooltip was shifted
            const arrowLeft = rect.left + rect.width / 2 - left;
            arrow.style.left = arrowLeft + 'px';
            arrowBorder.style.left = arrowLeft + 'px';
            
        } else if (spaceBottom > tooltipRect.height + arrowSize + margin) {
            // Position below
            position = 'bottom';
            top = rect.bottom + arrowSize;
            left = rect.left + rect.width / 2 - tooltipRect.width / 2;
            
            // Adjust horizontal position if needed
            if (left < margin) {
                left = margin;
            } else if (left + tooltipRect.width > viewportWidth - margin) {
                left = viewportWidth - tooltipRect.width - margin;
            }
            
            // Add arrow pointing up
            const arrowBorder = document.createElement('div');
            arrowBorder.className = 'tooltip-arrow-border arrow-up';
            tooltip.appendChild(arrowBorder);
            
            const arrow = document.createElement('div');
            arrow.className = 'tooltip-arrow arrow-up';
            tooltip.appendChild(arrow);
            
            // Adjust arrow position if tooltip was shifted
            const arrowLeft = rect.left + rect.width / 2 - left;
            arrow.style.left = arrowLeft + 'px';
            arrowBorder.style.left = arrowLeft + 'px';
            
        } else if (spaceRight > tooltipRect.width + arrowSize + margin && spaceRight > spaceLeft) {
            // Position to the right
            position = 'right';
            top = rect.top + rect.height / 2 - tooltipRect.height / 2;
            left = rect.right + arrowSize;
            
            // Adjust vertical position if needed
            if (top < margin) {
                top = margin;
            } else if (top + tooltipRect.height > viewportHeight - margin) {
                top = viewportHeight - tooltipRect.height - margin;
            }
            
            // Add arrow pointing left
            const arrowBorder = document.createElement('div');
            arrowBorder.className = 'tooltip-arrow-border arrow-left';
            tooltip.appendChild(arrowBorder);
            
            const arrow = document.createElement('div');
            arrow.className = 'tooltip-arrow arrow-left';
            tooltip.appendChild(arrow);
            
            // Adjust arrow position if tooltip was shifted
            const arrowTop = rect.top + rect.height / 2 - top;
            arrow.style.top = arrowTop + 'px';
            arrowBorder.style.top = arrowTop + 'px';
            
        } else {
            // Position to the left
            position = 'left';
            top = rect.top + rect.height / 2 - tooltipRect.height / 2;
            left = rect.left - tooltipRect.width - arrowSize;
            
            // Adjust vertical position if needed
            if (top < margin) {
                top = margin;
            } else if (top + tooltipRect.height > viewportHeight - margin) {
                top = viewportHeight - tooltipRect.height - margin;
            }
            
            // Add arrow pointing right
            const arrowBorder = document.createElement('div');
            arrowBorder.className = 'tooltip-arrow-border arrow-right';
            tooltip.appendChild(arrowBorder);
            
            const arrow = document.createElement('div');
            arrow.className = 'tooltip-arrow arrow-right';
            tooltip.appendChild(arrow);
            
            // Adjust arrow position if tooltip was shifted
            const arrowTop = rect.top + rect.height / 2 - top;
            arrow.style.top = arrowTop + 'px';
            arrowBorder.style.top = arrowTop + 'px';
        }
        
        // Apply position
        tooltip.style.top = top + 'px';
        tooltip.style.left = left + 'px';
        tooltip.className = 'tooltip-content active position-' + position;
    }
    
    async function showTooltip(element, word) {
        // Clear any pending hide timeout
        if (tooltipTimeout) {
            clearTimeout(tooltipTimeout);
            tooltipTimeout = null;
        }
        
        // If it's the same word, just keep showing
        if (currentWord === word && currentTooltip === element) {
            return;
        }
        
        // Hide current tooltip if different element
        if (currentTooltip && currentTooltip !== element) {
            hideTooltip(true); // immediate hide for different word
        }
        
        currentWord = word;
        currentTooltip = element;
        
        // Check if tooltip already exists
        let tooltip = document.querySelector('#tooltip-container');
        if (!tooltip) {
            // Create tooltip container (only one for the entire page)
            tooltip = document.createElement('div');
            tooltip.id = 'tooltip-container';
            tooltip.className = 'tooltip-content';
            document.body.appendChild(tooltip);
            
            // Prevent tooltip from closing when hovering over it
            tooltip.addEventListener('mouseenter', (e) => {
                e.stopPropagation();
                if (tooltipTimeout) {
                    clearTimeout(tooltipTimeout);
                    tooltipTimeout = null;
                }
            });
            
            tooltip.addEventListener('mouseleave', (e) => {
                e.stopPropagation();
                tooltipTimeout = setTimeout(() => hideTooltip(false), 300);
            });
            
            // Prevent clicks on tooltip from propagating
            tooltip.addEventListener('click', (e) => {
                e.stopPropagation();
            });
        }
        
        // Set loading state
        tooltip.innerHTML = '<div class="loading">Loading...</div>';
        tooltip.classList.add('active');
        
        // Position tooltip initially
        positionTooltip(tooltip, element);
        
        // Fetch data
        const data = await fetchTooltipData(word);
        
        if (data) {
            // Build tooltip content
            let html = `<div class="tooltip-description">${data.description}</div>`;
            html += '<div class="tooltip-links">';
            for (const link of data.links) {
                html += `<a href="${link.url}" target="_blank" class="tooltip-link">${link.text}</a>`;
            }
            html += '</div>';
            tooltip.innerHTML = html;
        } else {
            tooltip.innerHTML = '<div class="loading">No information available</div>';
        }
        
        // Reposition after content is loaded (size might have changed)
        positionTooltip(tooltip, element);
    }
    
    function hideTooltip(immediate = false) {
        const tooltip = document.querySelector('#tooltip-container');
        if (tooltip) {
            tooltip.classList.remove('active');
        }
        currentTooltip = null;
        currentWord = null;
    }
    
    function initTooltips() {
        // Prevent re-initialization causing infinite loops
        if (initialized) {
            return;
        }
        
        // Find all tooltip words and add event listeners
        document.querySelectorAll('.tooltip-word').forEach(element => {
            // Check if already initialized
            if (element.hasAttribute('data-tooltip-initialized')) {
                return;
            }
            element.setAttribute('data-tooltip-initialized', 'true');
            
            const word = element.dataset.word;
            
            element.addEventListener('mouseenter', (e) => {
                // Clear any pending hide
                if (tooltipTimeout) {
                    clearTimeout(tooltipTimeout);
                    tooltipTimeout = null;
                }
                // Add small delay before showing to prevent flicker
                if (showTimeout) {
                    clearTimeout(showTimeout);
                }
                showTimeout = setTimeout(() => {
                    showTooltip(element, word);
                }, 100);
            });
            
            element.addEventListener('mouseleave', (e) => {
                // Clear show timeout if leaving before tooltip shows
                if (showTimeout) {
                    clearTimeout(showTimeout);
                    showTimeout = null;
                }
                
                // Check if we're moving to the tooltip itself
                const relatedTarget = e.relatedTarget;
                if (relatedTarget && (relatedTarget.id === 'tooltip-container' || relatedTarget.closest('#tooltip-container'))) {
                    // Moving to tooltip, don't hide
                    return;
                }
                
                // Add delay before hiding to allow moving to tooltip
                tooltipTimeout = setTimeout(() => hideTooltip(false), 300);
            });
            
            // Handle mousemove to ensure tooltip stays visible
            element.addEventListener('mousemove', (e) => {
                if (currentWord === word && currentTooltip === element) {
                    // Already showing this tooltip, clear any hide timeout
                    if (tooltipTimeout) {
                        clearTimeout(tooltipTimeout);
                        tooltipTimeout = null;
                    }
                }
            });
        });
        
        initialized = true;
    }
    
    // Initialize tooltips when DOM is ready
    if (document.readyState === 'loading') {
        document.addEventListener('DOMContentLoaded', initTooltips);
    } else {
        // DOM is already loaded
        setTimeout(initTooltips, 100);
    }
    
    // Only watch for new content, not re-initialize everything
    const observer = new MutationObserver((mutations) => {
        mutations.forEach((mutation) => {
            mutation.addedNodes.forEach((node) => {
                if (node.nodeType === 1) { // Element node
                    // Check if it's a tooltip word or contains tooltip words
                    if (node.classList && node.classList.contains('tooltip-word')) {
                        if (!node.hasAttribute('data-tooltip-initialized')) {
                            initSingleTooltip(node);
                        }
                    } else if (node.querySelectorAll) {
                        node.querySelectorAll('.tooltip-word').forEach(el => {
                            if (!el.hasAttribute('data-tooltip-initialized')) {
                                initSingleTooltip(el);
                            }
                        });
                    }
                }
            });
        });
    });
    
    function initSingleTooltip(element) {
        if (element.hasAttribute('data-tooltip-initialized')) {
            return;
        }
        element.setAttribute('data-tooltip-initialized', 'true');
        
        const word = element.dataset.word;
        
        element.addEventListener('mouseenter', (e) => {
            if (tooltipTimeout) {
                clearTimeout(tooltipTimeout);
                tooltipTimeout = null;
            }
            if (showTimeout) {
                clearTimeout(showTimeout);
            }
            showTimeout = setTimeout(() => {
                showTooltip(element, word);
            }, 100);
        });
        
        element.addEventListener('mouseleave', (e) => {
            if (showTimeout) {
                clearTimeout(showTimeout);
                showTimeout = null;
            }
            
            const relatedTarget = e.relatedTarget;
            if (relatedTarget && (relatedTarget.id === 'tooltip-container' || relatedTarget.closest('#tooltip-container'))) {
                return;
            }
            
            tooltipTimeout = setTimeout(() => hideTooltip(false), 300);
        });
        
        element.addEventListener('mousemove', (e) => {
            if (currentWord === word && currentTooltip === element) {
                if (tooltipTimeout) {
                    clearTimeout(tooltipTimeout);
                    tooltipTimeout = null;
                }
            }
        });
    }
    
    // Start observing after initialization
    setTimeout(() => {
        observer.observe(document.body, { childList: true, subtree: true });
    }, 500);
    
    // Reposition tooltip on window resize/scroll
    window.addEventListener('resize', () => {
        if (currentTooltip) {
            const tooltip = document.querySelector('#tooltip-container');
            if (tooltip && tooltip.classList.contains('active')) {
                positionTooltip(tooltip, currentTooltip);
            }
        }
    });
    
    window.addEventListener('scroll', () => {
        if (currentTooltip) {
            const tooltip = document.querySelector('#tooltip-container');
            if (tooltip && tooltip.classList.contains('active')) {
                positionTooltip(tooltip, currentTooltip);
            }
        }
    });
</script>
'''


from nicegui import ui, app, run
import asyncio, datetime, re, os
from biblemateweb.pages.ai.stream import stream_response
from biblemateweb import BIBLEMATEWEB_APP_DIR, get_translation, markdown2html, get_watermark, DEFAULT_MESSAGES
from biblemateweb.mcp_tools.elements import TOOL_ELEMENTS
from biblemateweb.mcp_tools.tools import TOOLS
from biblemateweb.api.api import get_api_content
from biblemate.core.systems import get_system_tool_instruction
from copy import deepcopy
from agentmake import readTextFile

def chapter2verses(request:str) -> str:
    return re.sub("[Cc][Hh][Aa][Pp][Tt][Ee][Rr] ([0-9]+?)([^0-9])", r"\1:1-180\2", request)

def ai_chat(gui=None, q="", **_):
    
    START = True

    DOWNLOAD_CONTAINER = None
    SEND_BUTTON = None
    REQUEST_INPUT = None
    MESSAGES = None
    MESSAGE_CONTAINER = None
    SCROLL_AREA = None
    CANCEL_EVENT = None
    CANCEL_NOTIFICATION = None
    DELETE_DIALOG = None

    SYSTEM_TOOL_SELECTION = readTextFile(os.path.join(BIBLEMATEWEB_APP_DIR, "mcp_tools", "system_tool_selection_lite.md"))

    TOOL_INSTRUCTION_PROMPT = """Please transform the following suggestions into clear, precise, and actionable instructions."""
    TOOL_INSTRUCTION_SUFFIX = """

# Remember

* Provide me with the instructions directly.
* Do not start your response, like, 'Here are the insturctions ...'
* Do not ask me if I want to execute the instruction."""

    def reset_ui():
        nonlocal SEND_BUTTON, REQUEST_INPUT, CANCEL_EVENT, CANCEL_NOTIFICATION, MESSAGES, DOWNLOAD_CONTAINER
        if not CANCEL_EVENT.is_set():
            CANCEL_EVENT.set()
        if CANCEL_EVENT is not None:
            CANCEL_EVENT = None
        SEND_BUTTON.enable()
        SEND_BUTTON.set_text(get_translation("Send"))
        SEND_BUTTON.props('color=primary')
        REQUEST_INPUT.enable()
        REQUEST_INPUT.set_value("")
        REQUEST_INPUT.run_method('focus')
        if CANCEL_NOTIFICATION is not None:
            CANCEL_NOTIFICATION.dismiss()
            CANCEL_NOTIFICATION = None
        with MESSAGE_CONTAINER:
            with ui.row().classes('w-full justify-center') as DOWNLOAD_CONTAINER:
                ui.button(get_translation("Download the whole Conversation"), on_click=lambda: download_all_content(MESSAGES))

    async def stop_confirmed():
        nonlocal DELETE_DIALOG, CANCEL_NOTIFICATION, CANCEL_EVENT, SEND_BUTTON
        SEND_BUTTON.disable()
        DELETE_DIALOG.close()
        CANCEL_NOTIFICATION = ui.notification(get_translation("Stopping..."), timeout=None, spinner=True)
        if CANCEL_EVENT is not None and not CANCEL_EVENT.is_set():
            CANCEL_EVENT.set()
        await asyncio.sleep(0)

    def download_all_content(messages):
        messages_copy = deepcopy(messages)
        content = """---

I'm BibleMate AI, an autonomous agent designed to assist you with your Bible study.

---

"""
        content += "\n\n---\n\n".join([("[REQUEST] "+i.get("content", "")) if index%2 == 0 else ("[RESPONSE] "+i.get("content", "")) for index, i in enumerate(messages_copy[1:])])
        content += get_watermark()
        ui.download(content.encode('utf-8'), 'BibleMate_AI_Conversation.txt')
        ui.notify(get_translation("Downloaded!"), type='positive')

    async def handle_send_click():

        """Handles the logic when the Send button is pressed."""
        nonlocal REQUEST_INPUT, SCROLL_AREA, MESSAGE_CONTAINER, SEND_BUTTON, MESSAGES, CANCEL_EVENT, START, DOWNLOAD_CONTAINER, DELETE_DIALOG

        if DOWNLOAD_CONTAINER is not None:
            DOWNLOAD_CONTAINER.clear()

        if START:
            START = False
            MESSAGE_CONTAINER.clear()

        if CANCEL_EVENT is None or CANCEL_EVENT.is_set():
            CANCEL_EVENT = asyncio.Event() # do not use threading.Event() in this case
        else:
            DELETE_DIALOG.open()
            return None

        if not MESSAGES:
            MESSAGES = deepcopy(DEFAULT_MESSAGES)
        prompt_markdown = None
        output_markdown = None
        if user_request := REQUEST_INPUT.value:
            gui.update_active_area2_tab_records(q=user_request)

            REQUEST_INPUT.disable()
            SEND_BUTTON.set_text(get_translation("Stop"))
            SEND_BUTTON.props('color=negative')

            with MESSAGE_CONTAINER:
                ui.chat_message(user_request,
                    #name='Eliran Wong',
                    stamp=datetime.datetime.now().strftime("%H:%M"),
                    avatar=app.storage.user['avatar'] if app.storage.user['avatar'].strip() else 'https://avatars.githubusercontent.com/u/25262722?s=96&v=4',
                )

                # when prompt-engineering is enabled
                if app.storage.user["prompt_engineering"]:
                    with ui.expansion(get_translation("Prompt Engineering"), icon='tune', value=True) \
                            .classes('w-full border rounded-lg shadow-sm') \
                            .props('header-class="font-bold text-lg text-secondary"') as prompt_expansion:
                        prompt_markdown = ui.markdown().style('font-size: 1.1rem')
                    user_request = await stream_response(MESSAGES, user_request, prompt_markdown, CANCEL_EVENT, system="improve_prompt_2", scroll_area=SCROLL_AREA)
                    if not user_request or user_request.strip() == "[NO_CONTENT]":
                        reset_ui()
                        return None
                    else:
                        # refine response
                        if "```" in user_request:
                            user_request = re.sub(r"^.*?(```improved_prompt|```)(.+?)```.*?$", r"\2", user_request, flags=re.DOTALL).strip()
                        # apply the last fix from stream output
                        prompt_markdown.content = user_request
                        await asyncio.sleep(0)
                        # close prompt expansion
                        prompt_expansion.close()
                
                # tool selection
                selected_tool = "get_direct_text_response" # default
                if app.storage.user["auto_tool_selection"]:
                    with ui.expansion(get_translation("Tool Selection"), icon='handyman', value=True) \
                            .classes('w-full border rounded-lg shadow-sm') \
                            .props('header-class="font-bold text-lg text-secondary"') as tools_expansion:
                        tools_markdown = ui.markdown().style('font-size: 1.1rem')
                    suggested_tools = await stream_response(MESSAGES, user_request, tools_markdown, CANCEL_EVENT, system=SYSTEM_TOOL_SELECTION, scroll_area=SCROLL_AREA)
                    if not suggested_tools or suggested_tools.strip() == "[NO_CONTENT]":
                        suggested_tools = ["get_direct_text_response"]
                    else:
                        # refine response
                        suggested_tools = re.sub(r"^.*?(\[.*?\]).*?$", r"\1", suggested_tools, flags=re.DOTALL)
                        try:
                            suggested_tools = eval(suggested_tools.replace("`", "'")) if suggested_tools.startswith("[") and suggested_tools.endswith("]") else ["get_direct_text_response"] # fallback to direct response
                        except:
                            suggested_tools = ["get_direct_text_response"]
                        # close prompt expansion
                        tools_expansion.close()

                    selected_tool = suggested_tools[0]
                    if not selected_tool in TOOLS:
                        selected_tool = "get_direct_text_response"
                    suggested_tools = "\n".join([f"{order+1}. `{i}`" for order, i in enumerate(suggested_tools)])
                    tools_markdown.content = f"## Suggested Tools\n\n{suggested_tools}\n\n## Selected Tool:\n\n`{selected_tool}`"
                    await asyncio.sleep(0)
                    selected_tool_description = TOOLS.get(selected_tool, "No description available.")
                    tool_instruction_draft = TOOL_INSTRUCTION_PROMPT + "\n\n# Suggestions\n\n"+user_request+f"\n\n# Tool Description of `{selected_tool}`\n\n"+selected_tool_description+TOOL_INSTRUCTION_SUFFIX
                    system_tool_instruction = get_system_tool_instruction(selected_tool, selected_tool_description)
                    with ui.expansion(get_translation("Tool Instruction"), icon='handyman', value=True) \
                            .classes('w-full border rounded-lg shadow-sm') \
                            .props('header-class="font-bold text-lg text-secondary"') as tool_instruction_expansion:
                        tool_instruction_markdown = ui.markdown().style('font-size: 1.1rem')
                    user_request = await stream_response(MESSAGES, tool_instruction_draft, tool_instruction_markdown, CANCEL_EVENT, system=system_tool_instruction, scroll_area=SCROLL_AREA)
                    if not user_request or user_request.strip() == "[NO_CONTENT]":
                        reset_ui()
                        return None
                    else:
                        # apply the last fix from stream output
                        tool_instruction_markdown.content = user_request
                        await asyncio.sleep(0)
                        # close prompt expansion
                        tool_instruction_expansion.close()

                # generate response
                n = None
                answers = None
                try:
                    if selected_tool == "get_direct_text_response":
                        if app.storage.user["use_agent"]:
                            with ui.expansion(get_translation("Agent"), icon='support_agent', value=True) \
                                    .classes('w-full border rounded-lg shadow-sm') \
                                    .props('header-class="font-bold text-lg text-secondary"') as agent_expansion:
                                agent_markdown = ui.markdown().style('font-size: 1.1rem')
                            output_markdown = ui.markdown().style('font-size: 1.1rem')
                            answers = await stream_response(MESSAGES, user_request, output_markdown, CANCEL_EVENT, system="auto", scroll_area=SCROLL_AREA, agent_expansion=agent_expansion, agent_markdown=agent_markdown)
                        else:
                            output_markdown = ui.markdown().style('font-size: 1.1rem')
                            answers = await stream_response(MESSAGES, user_request, output_markdown, CANCEL_EVENT, system=None, scroll_area=SCROLL_AREA)
                        # update
                        if not answers or answers.strip() == "[NO_CONTENT]":
                            reset_ui()
                            return None
                        else:
                            # apply the last fix from stream output
                            output_markdown.content = answers
                            await asyncio.sleep(0)
                    else:
                        element = TOOL_ELEMENTS.get(selected_tool)
                        # API access
                        if isinstance(element, str):
                            output_markdown = ui.markdown().style('font-size: 1.1rem')
                            n = ui.notification(get_translation("Loading..."), timeout=None, spinner=True)
                            await asyncio.sleep(0)
                            if not selected_tool == "search_the_whole_bible":
                                user_request = chapter2verses(user_request)
                            api_query = f"{element}{user_request}"
                            answers = await run.io_bound(get_api_content, api_query, app.storage.user["ui_language"], app.storage.client["custom"])
                            output_markdown.content = answers if answers else "[NO_CONTENT]"
                            await asyncio.sleep(0)
                        # AI generation
                        elif isinstance(element, dict):
                            system = element.pop("system") if "system" in element else None
                            if system == "auto":
                                with ui.expansion(get_translation("Agent"), icon='support_agent', value=True) \
                                        .classes('w-full border rounded-lg shadow-sm') \
                                        .props('header-class="font-bold text-lg text-secondary"') as agent_expansion:
                                    agent_markdown = ui.markdown().style('font-size: 1.1rem')
                            else:
                                agent_markdown = None
                                agent_expansion = None
                            output_markdown = ui.markdown().style('font-size: 1.1rem')
                            answers = await stream_response(MESSAGES, user_request, output_markdown, CANCEL_EVENT, system=system, scroll_area=SCROLL_AREA, agent_expansion=agent_expansion, agent_markdown=agent_markdown, **element)
                            # update
                            if not answers or answers.strip() == "[NO_CONTENT]":
                                reset_ui()
                                return None
                            else:
                                # apply the last fix from stream output
                                output_markdown.content = answers
                                await asyncio.sleep(0)
                except Exception as e:
                    output_markdown.content = f"[{get_translation('Error')}: {str(e)}]"
                    await asyncio.sleep(0)
                    #import traceback
                    #traceback.print_exc()
                finally:
                    if n is not None:
                        n.spinner = False
                        n.dismiss()
                        n = None
                    if answers and not answers.strip() == "[NO_CONTENT]":
                        MESSAGES += [
                            {"role": "user", "content": user_request},
                            {"role": "assistant", "content": answers},
                        ]

                #reset
                reset_ui()

    with ui.column().classes('w-full h-screen no-wrap gap-0') as chat_container:

        def check_screen(ev):
            nonlocal gui, chat_container
            if gui.is_portrait and app.storage.user["layout"] == 2:
                chat_container.classes(remove='h-screen', add='h-[50vh]')
            else:
                chat_container.classes(remove='h-[50vh]', add='h-screen')
            chat_container.update()

        # check screen when loaded
        check_screen(None)
        # bind
        ui.on('resize', check_screen)

        # Capture the ui.scroll_area instance in the global variable
        # w-full flex-grow p-4 border border-gray-300 rounded-lg mb-2
        with ui.column().classes('w-full flex-grow overflow-hidden'):
            with ui.scroll_area().classes('w-full p-4 border rounded-lg h-full') as SCROLL_AREA:
                with ui.column().classes('w-full items-start gap-2') as MESSAGE_CONTAINER:
                    welcome_message = """**Hello! I’m BibleMate AI.** You’ve enabled my Chat Mode capabilities.

I’m here to chat with you about anything in the Bible. Whether you have questions about a verse, need historical context, or just want to explore a topic, let’s talk!

What would you like to discuss today? Enter your message below to get started.

---

💡 Tips — Enhance your experience:

* Enhance: Check this to automatically improve and clarify your prompt.
* Agent: Check this to engage my advanced reasoning for more detailed responses.
* Tools: Check this to let me access my dedicated suite of Bible study tools for better accuracy.

---"""
                    ui.chat_message(markdown2html(get_translation(welcome_message if app.storage.user["ui_language"] == "eng" else get_translation("welcome_chat_mode"))),
                        name='BibleMate AI',
                        stamp=datetime.datetime.now().strftime("%H:%M"),
                        avatar='https://avatars.githubusercontent.com/u/25262722?s=96&v=4',
                        text_html=True,
                        sanitize=False,
                    )
                    ui.markdown(f"---\n\n## {get_translation('Other AI Modes')}\n\n")
                    ui.button(get_translation("Partner Mode"), on_click=lambda: gui.load_area_2_content(title="partner"))
                    ui.button(get_translation("Agent Mode"), on_click=lambda: gui.load_area_2_content(title="chat"))
                    ui.markdown("---")

        with ui.row().classes('w-full flex-nowrap items-end mb-0'):
            REQUEST_INPUT = ui.textarea(placeholder=get_translation("Enter your request here...")).props('rows=4').classes('flex-grow h-full resize-none').on('keydown.shift.enter.prevent', handle_send_click)
            with ui.column().classes('h-full justify-between gap-2'):
                ui.checkbox(get_translation("Auto-scroll")).classes('w-full').bind_value(app.storage.user, 'auto_scroll').props('dense').tooltip(get_translation("Scroll to the end automatically"))
                SEND_BUTTON = ui.button('Send', on_click=handle_send_click).classes('w-full')
        
        with ui.row().classes('w-full flex-nowrap items-end mb-27'):
            ui.checkbox(get_translation("Enhance")).classes('w-full').bind_value(app.storage.user, 'prompt_engineering').props('dense').tooltip(get_translation("Improve Prompt"))
            ui.checkbox(get_translation("Agent")).classes('w-full').bind_value(app.storage.user, 'use_agent').props('dense').tooltip(get_translation("Create agents to improve responses"))
            ui.checkbox(get_translation("Tools")).classes('w-full').bind_value(app.storage.user, 'auto_tool_selection').props('dense').tooltip(get_translation("Use tools to improve responses"))

        ui.label('BibleMate AI | © 2025-2026 | Eliran Wong')

        # Dialog to confirm the reset
        with ui.dialog() as DELETE_DIALOG, ui.card():
            ui.label(f'{get_translation("Stop running task?")}?')
            with ui.row().classes('justify-end w-full'):
                ui.button(get_translation("Cancel"), on_click=DELETE_DIALOG.close).props('flat text-color=secondary')
                ui.button(get_translation("Stop"), color='negative', on_click=stop_confirmed)

        if q:
            REQUEST_INPUT.set_value(q)
        REQUEST_INPUT.run_method('focus')
        ui.run_javascript(f'''
            const el = document.getElementById({REQUEST_INPUT.id}).querySelector('textarea');
            el.setSelectionRange(el.value.length, el.value.length);
        ''')



import sqlite3, os
import base64
from typing import Optional
from biblemateweb import BIBLEMATEWEB_DATA

def get_image_data_uri(module_table: str, image_path_key: str, mime_type: str = "image/png") -> Optional[str]:
    """
    Retrieves image BLOB from SQLite, encodes it to base64, and formats it as a Data URI.

    Args:
        module_table (str): Path to the SQLite table.
        image_path_key (str): The unique 'path' value (e.g., file name) to retrieve.
        mime_type (str): The MIME type of the stored image (e.g., 'image/png', 'image/jpeg').

    Returns:
        Optional[str]: The complete Data URI string, or None if the image is not found.
    """
    db_path = os.path.join(BIBLEMATEWEB_DATA, "images.sqlite")
    conn = None
    try:
        # 1. Connect to the database
        conn = sqlite3.connect(db_path)
        cursor = conn.cursor()

        # 2. Retrieve the BLOB data
        # Use a parameterized query to prevent SQL injection
        cursor.execute(f"SELECT image FROM {module_table} WHERE path = ?", (image_path_key,))
        
        row = cursor.fetchone()
        
        if row is None:
            print(f"Error: No image found with path: {image_path_key}")
            return None
            
        # The BLOB data is the first (and only) element in the row tuple
        image_blob = row[0]
        
        # 3. Encode the BLOB data to base64
        # b64encode converts binary data to a bytes-like object containing base64 data
        base64_bytes = base64.b64encode(image_blob)
        
        # 4. Decode the bytes to a string for use in HTML
        base64_string = base64_bytes.decode('utf-8')
        
        # 5. Format as a Data URI
        # The format is: data:[<MIME-type>];base64,<data>
        data_uri = f"data:{mime_type};base64,{base64_string}"
        
        return data_uri

    except sqlite3.Error as e:
        print(f"SQLite error: {e}")
        return None
    finally:
        if conn:
            conn.close()


from agentmake import agentmake, DEFAULT_AI_BACKEND, unpack_instruction_content, unpack_system_content
from agentmake.utils.text_wrapper import get_stream_event_text
from agentmake.utils.read_assistant_response import is_openai_style
from nicegui import ui, app, run
import asyncio
from biblemateweb import config, get_translation, DEFAULT_MESSAGES


async def stream_response(messages, user_request, response_markdown, cancel_event, system=None, scroll_area=None, agent_expansion=None, agent_markdown=None, **kwargs):
    def get_next_chunk(iterator):
        """
        Runs in a separate thread. 
        Returns the next item, or None if the iterator is exhausted.
        """
        try:
            return next(iterator)
        except StopIteration:
            return None
        except Exception as e:
            return e  # Return the error to be handled in the main loop

    if "instruction" in kwargs:
        instruction_content = kwargs.pop("instruction")
        instruction_content = unpack_instruction_content(instruction_content)
        # refine user request
        user_request = instruction_content + "\n" + user_request

    if system == "auto":
        system = await stream_response(DEFAULT_MESSAGES, user_request, agent_markdown, cancel_event, system="bible/create_agent", scroll_area=scroll_area)
        if not system or system.strip() == "[NO_CONTENT]":
            return None
        else:
            # refine response
            system = system.replace("should:", "should:\n")
            system = system.replace("examples:", "examples:\n")
            if system.startswith("```agent\n"):
                system = system[9:]
            if system.endswith("```"):
                system = system[:-3].strip()
            agent_markdown.content = system
            await asyncio.sleep(0)
            # close prompt expansion
            if agent_expansion is not None:
                agent_expansion.close()
    elif system is not None:
        system_content = unpack_system_content(system)
        system = None
        # refine user request
        # Note: when a conversation goes long, system message placed in the beginning of the messages chain does not work properly.
        # Embed the system message into the user request to enhance the response
        user_request = f"""---

START OF YOUR NEW ROLE

---

{system_content}

---

START OF MY REQUEST

---

{user_request}"""

    # get streaming object
    n = ui.notification(get_translation("Loading..."), timeout=None, spinner=True)
    if app.storage.user["auto_scroll"] and scroll_area is not None:
        # Give the client a moment to render the new content
        #await asyncio.sleep(0.1)
        # scroll to the bottom
        await ui.run_javascript(f'getElement({scroll_area.id}).setScrollPosition("vertical", 99999, 300)')
    await asyncio.sleep(0)
    # run completion
    text_chunks = ""
    completion = await run.io_bound(
        agentmake, 
        messages, 
        system=system,
        backend=app.storage.user["ai_backend"] if app.storage.user["api_key"].strip() and app.storage.user["ai_backend"] else config.ai_backend if config.ai_backend else DEFAULT_AI_BACKEND, 
        model=app.storage.user["ai_model"].strip() if app.storage.user["api_key"].strip() and app.storage.user["ai_model"].strip() else None, 
        api_key=app.storage.user["api_key"].strip() if app.storage.user["api_key"].strip() else None,
        api_endpoint=app.storage.user["api_endpoint"].strip() if app.storage.user["api_key"].strip() and app.storage.user["api_endpoint"].strip() else None,
        max_tokens=app.storage.user["max_tokens"] if app.storage.user["max_tokens"] and app.storage.user["api_endpoint"].strip() else None,
        temperature=app.storage.user["temperature"] if app.storage.user["temperature"] and app.storage.user["api_endpoint"].strip() else None,
        follow_up_prompt=user_request, 
        stream=True, 
        print_on_terminal=False, 
        stream_events_only=True, 
        #streaming_event=cancel_event, # this works only for unpacking completion text; not useful in this case
        **kwargs,
    )
    n.message = get_translation("Running...")
    await asyncio.sleep(0)
    try:
        while cancel_event is not None and not cancel_event.is_set():
            # 0. SAFETY CHECK: Stop if user closed the tab
            if not ui.context.client.connected:
                #print("Client disconnected. Stopping stream.")
                break
            # 1. Fetch next chunk using the helper
            # This moves the blocking 'next()' call AND the StopIteration catch 
            # into the thread. We just wait for the result.
            event = await run.io_bound(get_next_chunk, completion)

            # 2. Explicit check: Is the iterator exhausted?
            if event is None:
                #print("Stream finished successfully.") # Debug log
                break
            # 3. Check for errors returned by the helper
            elif isinstance(event, Exception):
                ui.notify(f"Stream interrupted: {str(event)}", type='warning')
                break

            # --- Process the valid event below ---
            if text_chunk := get_stream_event_text(event, openai_style=is_openai_style(config.ai_backend if config.ai_backend else DEFAULT_AI_BACKEND)):
                text_chunks += text_chunk
                response_markdown.content = text_chunks
                await asyncio.sleep(0)
        # when cancelled
        if cancel_event is not None and cancel_event.is_set():
            n.message = get_translation("Cancelled!")
            response_markdown.content = f"[{get_translation("Cancelled!")}]"
            await asyncio.sleep(0)
        # when done
        else:
            n.message = get_translation("Done!")
            if app.storage.user["auto_scroll"] and scroll_area is not None:
                # Give the client a moment to render the new content
                await asyncio.sleep(0.1)
                # scroll to the bottom
                await ui.run_javascript(f'getElement({scroll_area.id}).setScrollPosition("vertical", 99999, 300)')
    except Exception as e:
        response_markdown.content = f"[{get_translation('Error')}: {str(e)}]"
        await asyncio.sleep(0)
        # check traceback
        #import traceback
        #traceback.print_exc()
        cancel_event.set()

    # stop this spinner
    n.spinner = False
    n.dismiss()
    n = None
    if cancel_event.is_set():
        cancel_event = None
        return None
    
    return text_chunks.replace(" ", " ").replace("‑", "-") # replace non-standard characters in some LLM output, e.g. gpt-oss



translation_tc = {
    "Add Bible Tab": "新聖經標籤頁",
    "Add Parallel Tab": "新平行標籤頁",
    "Add Tool Tab": "新工具標籤頁",
    "Agent": "代理",
    "AI Chat": "對話模式",
    "AI Commentary": "AI 註釋",
    "Ask BibleMate": "問 BibleMate",
    "Bible Audio": "聖經語音",
    "Bible Chronology": "聖經時序",
    "Bible Commentaries": "聖經註釋",
    "Bible Only": "只顯示聖經",
    "Bible Parallels": "平行經文",
    "Bible Podcast": "聖經廣播",
    "Bible Promises": "聖經應許",
    "Bible": "聖經",
    "Bibles": "聖經",
    "Parallel Bibles": "平行聖經",
    "Timelines": "聖經時間軸",
    "Bible Timelines": "聖經時間軸",
    "Characters": "聖經人物",
    "Close Bible Tab": "關閉聖經標籤頁",
    "Close Others": "關閉其他",
    "Close Parallel Tab": "關閉平行標籤頁",
    "Close Tool Tab": "關閉工具標籤頁",
    "Commentaries": "聖經註釋",
    "Comparison": "版本對比",
    "Copy": "複製",
    "Cross-references": "經文串珠",
    "Dictionaries": "聖經字典",
    "Discourse Analysis": "語法分析",
    "Encyclopedias": "百科全書",
    "Go": "開啓",
    "Indexes": "資源索引",
    "Chapter Indexes": "全章索引",
    "Verse Indexes": "單節索引",
    "Lexicons": "原文詞典",
    "Locations": "聖經地點",
    "Maps": "聖經地圖",
    "Morphological Data": "形態學數據",
    "Morphology": "形態學分析",
    "Names": "聖經名字",
    "Parallels": "平行經文",
    "Partner Mode": "搭檔模式",
    "Preferences": "偏好設定",
    "Promises": "聖經應許",
    "Relationships": "人物關係",
    "Screen": "全螢幕",
    "Fullscreen": "全螢幕",
    "Swap": "版面轉換",
    "Sync": "同步",
    "Tool Only": "只顯示工具",
    "Topics": "聖經主題",
    "Treasury": "經文寶庫",
    "Treasury of Scripture Knowledge": "經文寶庫",
    "Bible & Tool": "聖經與工具",
    "Light Mode": "淺色模式",
    "Dark Mode": "深色模式",
    "Verses": "搜索經文",
    "Tools": "工具",
    "Search": "搜索",
    "Quick search": "快速搜索",
    "AI": "人工智能",
    "Settings": "設定",
    "Swap Layout": "版面轉換",
    "Select Bible Verse": "聖經選單",
    "Toggle Display of Swap Layout Button": "[開啓/關閉] 版面轉換按鈕",
    "Toggle Display of Bible Selection Button": "[開啓/關閉] 聖經選單按鈕",
    "Toggle Bible Synchronization": "[開啓/關閉] 聖經同步功能",
    "Toggle Fullscreen": "[開啓/關閉] 全螢幕",
    "Prev Chapter": "上一章",
    "Next Chapter": "下一章",
    "OT": "舊約",
    "NT": "新約",
    "Select Book": "選擇書卷",
    "Select Chapter": "選擇章數",
    "Select Verse": "選擇節數",
    "Back": "返回",
    "Share": "分享",
    "Note": "筆記",
    "Notes": "筆記",
    "Bible Notes": "聖經筆記",
    "Notepad": "筆記本",
    "Add Note": "新增筆記",
    "Edit Note": "編輯筆記",
    "Save": "儲存",
    "Download": "下載",
    "Import": "匯入",
    "Delete": "刪除",
    "Logout": "登出",
    "Read": "閱讀",
    "Edit": "編輯",
    "Edit Request": "編輯需求",
    "Edit Plan": "編輯計劃",
    "Edit Tools": "編輯工具",
    "Edit Conversation": "編輯對話",
    "Edit Rounds": "編輯回合",
    "Trim Rounds": "修剪回合",
    "Trim Rounds from:": "從下面選擇的回合開始修剪：",
    "Clear": "清除",
    "Download Index": "下載索引",
    "Rebuild Index": "重建索引",
    "Download All": "下載所有筆記",
    "Book Note": "書卷筆記",
    "Chapter Note": "全章筆記",
    "Verse Note": "經文筆記",
    "Book Tools": "書卷工具",
    "Chapter Tools": "全章工具",
    "Verse Tools": "單節工具",
    "Chapter Summary": "全章摘要",
    "Commentary": "經文註釋",
    "Book Analysis": "書卷分析",
    "Overview": "概述",
    "Structural Outline": "結構大綱",
    "Logical Flow": "書卷脈絡",
    "Historical Setting": "歷史背景",
    "Themes": "主題",
    "Keywords": "關鍵詞",
    "Theology": "神學",
    "Canonical Placement": "正典位置",
    "Practical Living": "生活應用",
    "Summary": "摘要",
    "Chapter Map": "全章地圖",
    "Loop": "循環",
    "Edit Mode": "編輯模式",
    "Read Mode": "閱讀模式",
    "Literal": "字面",
    "Regex": "模式",
    "Semantic": "語義",
    "Case-sensitive": "大小寫",
    "Type to filter": "輸入以篩選",
    "results": "結果",
    "result": "結果",
    "Search for words or refs": "搜索經文或參照（如：申 6:4; 約 3:16-18）",
    "Search Bible dictionaries": "搜索聖經字典",
    "Search for a bible topic": "搜索聖經主題",
    "Search for bible promises": "搜索聖經應許",
    "Search for parallel passages": "搜索平行經文",
    "Search for a name or meaning": "搜索名字或含義",
    "Search for location(s) or verse reference(s)": "搜索地點或經文參照",
    "Search for a bible location": "搜索聖經地點",
    "Search for a bible character": "搜索聖經人物",
    "Show Verses": "顯示所有經文",
    "Distance": "距離",
    "From": "從",
    "To": "到",
    "km": "公里",
    "miles": "英里",
    "Loading...": "載入中...",
    "Connecting...": "連線中...",
    "Canceling...": "取消中...",
    "Running...": "執行中...",
    "Stopping...": "停止中...",
    "Stop running task?": "停止執行中的工作？",
    "Enter bible verse reference(s) here...": "在此輸入經文參照...",
    "Start typing your notes here...": "在此開始輸入您的筆記...",
    "Reset All Preferences": "重置所有偏好設定",
    "Language": "語言",
    "Delete": "刪除",
    "Cancel": "取消",
    "Resume": "繼續",
    "Reset": "重置",
    "Confirm": "確認",
    "Partner Mode Review": "夥伴模式審查",
    "Cancelled!": "已取消！",
    "Done!": "完成！",
    "Stop": "停止",
    "Start": "開始",
    "Send": "發送",
    "Auto-scroll": "自動滾動",
    "Scroll to the end automatically": "自動滾動至結尾",
    "Improve Prompt": "優化指示",
    "Enhance": "優化指示",
    "Enhanced Prompt": "經優化的指示",
    "Tools": "工具",
    "Tool Selection": "選用工具",
    "Semantic Searches": "語義搜索",
    "Frequently Used Resources": "常用資源",
    "User & Custom Data": "用戶與自定數據",
    "Tabs Management": "標籤頁管理",
    "Study Plan": "研讀計劃",
    "Prompt Engineering": "優化用戶要求",
    "Tool Instruction": "工具指示",
    "Agent": "代理",
    "Create agents to improve responses": "創建代理優化回應",
    "Use tools to improve responses": "使用工具優化回應",
    "Tool Selection": "選用工具",
    "Final Report": "總結報告",
    "Download the Final Report Only": "只下載總結報告",
    "Download the whole Conversation": "下載完整對話",
    "Downloaded!": "已下載！",
    "Checking daily limit...": "檢查每日限額...",
    "Saving access note...": "記錄使用日期中...",
    "Wrapping up...": "結尾中...",
    "Round": "回合",
    "Appearance": "外觀",
    "Creating agent...": "創建代理中...",
    "Created!": "已創建！",
    "Error": "錯誤",
    "welcome_agent_mode": """**你好！我是 BibleMate AI。** 你已開啟「代理模式 (Agent Mode)」功能。

我能協助你深入查考聖經的內容，包括制定結構化的研經計畫，並跨多種聖經工具執行複雜的多步驟任務。無論是主題式搜尋還是深入的經文分析，我都隨時為你提供協助。

今天你想研習什麼呢？請在下方輸入你的需求以開始執行。

---

💡 提示： 勾選「優化」框，系統將自動完善並釐清你的需求，以獲得更理想的結果。

---""",
    "welcome_chat_mode": """**你好！我是 BibleMate AI。**你已開啟「對話模式 (Chat Mode)」功能。

讓我們一起討論聖經中的任何事物吧！無論你對經文有疑問、需要了解歷史背景，還是想探討特定主題，我隨時都能與你交流。

今天你想討論什麼呢？請在下方輸入你的訊息以開始對話。

---

💡 提示 — 選用功能：

* 優化 (Enhance)： 勾選此項以自動完善並釐清你的提問。
* 代理 (Agent)： 勾選此項以啟動進階邏輯，獲取更詳盡的回答。
* 工具 (Tools)： 勾選此項讓我調用專屬的研經工具組，以確保回覆更加精確。

---""",
    "welcome_partner_mode": """**你好！我是 BibleMate AI。你已開啟「夥伴模式 (Partner Mode)」功能。

在此模式下，我們將作為團隊協作。與代理模式一樣，我會處理複雜的多步驟研經任務，但你始終擁有主導權。你可以審查並調整總體規劃、核准工具選擇，並修改我執行每一步時的指令。

今天想一起研讀什麼呢？請在下方輸入你的需求，開始我們的協作。

---

💡 提示： 勾選「最佳化」框，系統將自動完善並明確你的需求，以獲得更理想的結果。

---""",
    "Agent Mode": "代理模式",
    "Partner Mode": "夥伴模式",
    "Chat Mode": "對話模式",
    "Other AI Modes": "其他模式",
    "Enter your request here...": "在此輸入你的需求",
}